import logging
from pathlib import Path

import typer
from cookiecutter.main import cookiecutter

from wanna.core.loggers.wanna_logger import get_logger

from .plugins.runner import PluginRunner
from .version import perform_check

logger = get_logger(__name__)

logging.getLogger("smart_open").setLevel(logging.ERROR)
logging.getLogger("google").setLevel(logging.ERROR)
logging.getLogger("kfp").setLevel(logging.ERROR)

runner = PluginRunner()
app = runner.app


@app.command(name="version", help="Print your current and latest available version")
def version():
    perform_check()


@app.command(name="init", help="Initiate a new wanna-ml project from template.")
def init(
    output_dir: Path = typer.Option(
        ...,
        "--output-dir",
        prompt="Where do you want to initiate your wanna-ml repository? (input '.' to use the current directory)",
        help="The output directory where wanna-ml repository will be created",
    ),
    template: str = typer.Option(
        ...,
        "--template",
        "-t",
        help="The git repository of the template you want to use",
    ),
    checkout: str | None = typer.Option(
        None,
        "--checkout",
        "-c",
        help="The branch, tag or commit to checkout after cloning the repository",
    ),
    directory: str | None = typer.Option(
        None,
        "--directory",
        "-d",
        help="The directory within the repository to use as the template",
    ),
    overwrite_if_exists: bool = typer.Option(
        False,
        "--overwrite-if-exists",
        help="Overwrite the contents of the output directory if it exists",
    ),
    no_input: bool = typer.Option(
        False,
        "--no-input",
        help="Do not prompt for parameters and only use cookiecutter.json file content",
    ),
):
    result_dir = cookiecutter(
        template=template,
        checkout=checkout,
        directory=directory,
        output_dir=output_dir,
        overwrite_if_exists=overwrite_if_exists,
        no_input=no_input,
    )
    logger.user_success(f"Repo initiated at {result_dir}")


def wanna():
    """
    Main entrypoint for wanna cli
    """
    app()


if __name__ == "__main__":
    wanna()



from typing import Annotated, Literal

from pydantic import (
    BaseModel,
    ConfigDict,
    EmailStr,
    Extra,
    Field,
    StringConstraints,
    field_validator,
    model_validator,
)

from wanna.core.models.base_instance import BaseInstanceModel
from wanna.core.models.gcp_components import GPU, Disk, VMImage
from wanna.core.utils import validators

WorkbenchName = Annotated[
    str,
    StringConstraints(
        min_length=3, max_length=63, pattern="^[a-z][a-z0-9-]*[a-z0-9]$", to_lower=True
    ),
]


class BucketMount(BaseModel, extra=Extra.forbid):
    bucket_name: str
    mount_path: str = "/gcs"

    model_config = ConfigDict(extra="forbid")

    _bucket_name = field_validator("bucket_name")(validators.validate_bucket_name)


class NotebookEnvironment(BaseModel, extra=Extra.forbid):
    vm_image: VMImage | None = None
    docker_image_ref: str | None = None

    model_config = ConfigDict(extra="forbid")

    _ = model_validator(mode="before")(validators.validate_only_one_must_be_set)


class BaseWorkbenchModel(BaseInstanceModel):
    name: WorkbenchName
    machine_type: str = "e2-standard-2"
    gpu: GPU | None = None
    data_disk: Disk | None = None
    subnet: str | None = None
    tensorboard_ref: str | None = None


class InstanceModel(BaseWorkbenchModel):
    """
    - `type` - [str] (optional) Type of the notebook, instance, for dispatching purposes.
    - `name`- [str] Custom name for this instance
    - `project_id` - [str] (optional) Overrides GCP Project ID from the `gcp_profile` segment
    - `zone` - [str] (optional) Overrides zone from the `gcp_profile` segment
    - `region` - [str] (optional) Overrides region from the `gcp_profile` segment
    - `labels`- [dict[str, str]] (optional) Custom labels to apply to this instance
    - `service_account` - [str] (optional) Overrides service account from the `gcp_profile` segment
    - `network` - [str] (optional) Overrides network from the `gcp_profile` segment
    - `tags`- [dict[str, str]] (optional) Tags to apply to this instance
    - `metadata`- [Optional[dict[str, Any]]] (optional) Custom metadata to apply to this instance
    - `owner` - [str] This can be either a single user email address and that would be the only one
      able to access the notebook. Or service account and then everyone who has the iam.serviceAccounts.actAs
      permission on the specified service account will be able to connect.
    - `machine_type` - [str] (optional) GCP Compute Engine machine type
    - `gpu`- [GPU] (optional) The hardware GPU accelerator used on this instance.
    - `data_disk` - [Disk] (optional) Data disk configuration to attach to this instance.
    - `kernels` - [list[str]] (optional) Custom kernels given as links to container registry
    - `tensorboard_ref` - [str] (optional) Reference to Vertex Experimetes
    - `subnet`- [str] (optional) Subnetwork of a given network
    - `internal_ip_only` - [bool] (optional) Public or private (default) IP address
    - `idle_shutdown_timeout` - [int] (optional) Time in minutes, between 10 and 1440, defaults to 720. If None,
        there is no idle shutdown
    - `post_startup_script` - [str] (optional) Path to a script that will be executed after the instance is started.
    - `post_startup_script_behavior` - [str] Defines the behavior of the post startup script.
        Documentation https://cloud.google.com/vertex-ai/docs/workbench/instances/manage-metadata
    - `environment_auto_upgrade` - [str] (optional) Cron schedule for environment auto-upgrade.
    - `delete_to_trash` - [bool] (optional) If true, the instance will be deleted to trash.
    - `report_health` - [bool] (optional) If true, the instance will report health to Cloud Monitoring
    """

    type: Literal["instance"] = "instance"
    zone: str
    owner: EmailStr | None = None
    boot_disk: Disk | None = None
    environment: NotebookEnvironment = NotebookEnvironment(vm_image=VMImage())
    no_public_ip: bool = True
    enable_dataproc: bool = False
    enable_ip_forwarding: bool = False
    no_proxy_access: bool = False
    enable_monitoring: bool = True
    idle_shutdown_timeout: int | None = Field(ge=10, le=1440, default=720)  # 12 hours
    collaborative: bool = False
    env_vars: dict[str, str] | None = None
    bucket_mounts: list[BucketMount] | None = None
    post_startup_script: str | None = None  # todo: add validation for existing object in bucket
    post_startup_script_behavior: Literal[
        "run_once", "run_every_start", "download_and_run_every_start"
    ] = "run_once"
    environment_auto_upgrade: str | None = None
    delete_to_trash: bool = False
    report_health: bool = True

    _environment_auto_upgrade = field_validator("environment_auto_upgrade")(
        validators.validate_cron_schedule
    )
    _machine_type = field_validator("machine_type")(validators.validate_machine_type)



import pathlib
from pathlib import Path

import typer

from wanna.cli.plugins.base_plugin import BasePlugin
from wanna.cli.plugins.common_options import (
    instance_name_option,
    profile_name_option,
    push_mode_option,
    version_option,
    wanna_file_option,
)
from wanna.core.deployment.models import PushMode
from wanna.core.utils.config_loader import load_config_from_yaml


class JobPlugin(BasePlugin):
    """
    Plugin for building and deploying training jobs.
    """

    def __init__(self) -> None:
        super().__init__()
        self.register_many(
            [
                self.build,
                self.push,
                (self.run, {"allow_extra_args": True, "ignore_unknown_options": True}),
                (
                    self.run_manifest,
                    {"allow_extra_args": True, "ignore_unknown_options": True},
                ),
                self.stop,
                self.report,
            ]
        )

    @staticmethod
    def build(
        file: Path = wanna_file_option,
        profile_name: str = profile_name_option,
        instance_name: str = instance_name_option("job", "build"),
    ) -> None:
        """
        Create a manifest based on the wanna-ml config that can be later pushed or run.
        """

        config = load_config_from_yaml(file, gcp_profile_name=profile_name)
        workdir = pathlib.Path(file).parent.resolve()

        # doing this import here speeds up the CLI app considerably
        from wanna.core.services.jobs import JobService

        job_service = JobService(config=config, workdir=workdir)
        job_service.build(instance_name)

    @staticmethod
    def push(
        file: Path = wanna_file_option,
        profile_name: str = profile_name_option,
        version: str = version_option(instance_type="job"),
        instance_name: str = instance_name_option("job", "push"),
        mode: PushMode = push_mode_option,
    ) -> None:
        """
        Build and push manifest to Cloud Storage.
        """
        config = load_config_from_yaml(file, gcp_profile_name=profile_name)
        workdir = pathlib.Path(file).parent.resolve()

        # doing this import here speeds up the CLI app considerably
        from wanna.core.services.jobs import JobService

        job_service = JobService(config=config, workdir=workdir, version=version, push_mode=mode)
        manifests = job_service.build(instance_name)
        job_service.push(manifests)

    @staticmethod
    def run(
        ctx: typer.Context,
        file: Path = wanna_file_option,
        profile_name: str = profile_name_option,
        version: str = version_option(instance_type="job"),
        instance_name: str = instance_name_option("job", "run"),
        hp_params: Path = typer.Option(
            None, "--hp-params", "-hp", help="Path to the params file in yaml format"
        ),
        sync: bool = typer.Option(False, "--sync", "-s", help="Runs the job in sync mode"),
    ) -> None:
        """
        Run the job as specified in wanna-ml config. This command puts together build, push and run-manifest steps.
        """

        args, command = JobPlugin._extract_job_overrides(ctx.args)
        config = load_config_from_yaml(file, gcp_profile_name=profile_name)
        workdir = pathlib.Path(file).parent.resolve()

        # doing this import here speeds up the CLI app considerably
        from wanna.core.services.jobs import JobService

        job_service = JobService(config=config, workdir=workdir, version=version)
        manifests = job_service.build(instance_name)
        job_service.push(manifests, local=False)
        JobService.run(
            [str(p) for p in manifests],
            sync=sync,
            hp_params=hp_params,
            command_override=command,
            args_override=args,
        )

    @staticmethod
    def run_manifest(
        ctx: typer.Context,
        manifest: str = typer.Option(None, "--manifest", "-v", help="Job deployment manifest"),
        hp_params: Path = typer.Option(
            None, "--hp-params", "-hp", help="Path to the params file in yaml format"
        ),
        sync: bool = typer.Option(False, "--sync", "-s", help="Runs the pipeline in sync mode"),
    ) -> None:
        """
        Run the job as specified in the wanna-ml manifest.
        """
        args, command = JobPlugin._extract_job_overrides(ctx.args)

        # doing this import here speeds up the CLI app considerably
        from wanna.core.services.jobs import JobService

        JobService.run(
            manifests=[manifest],
            sync=sync,
            hp_params=hp_params,
            command_override=command,
            args_override=args,
        )

    @staticmethod
    def stop(
        file: Path = wanna_file_option,
        profile_name: str = profile_name_option,
        instance_name: str = instance_name_option("job", "stop"),
    ) -> None:
        """
        Stop a running job.
        """
        config = load_config_from_yaml(file, gcp_profile_name=profile_name)
        workdir = pathlib.Path(file).parent.resolve()

        # doing this import here speeds up the CLI app considerably
        from wanna.core.services.jobs import JobService

        job_service = JobService(config=config, workdir=workdir)
        job_service.stop(instance_name)

    @staticmethod
    def report(
        file: Path = wanna_file_option,
        profile_name: str = profile_name_option,
        instance_name: str = instance_name_option("job", "report"),
    ) -> None:
        """
        Displays a link to the cost report per wanna_project and optionally per job name.
        """
        config = load_config_from_yaml(file, gcp_profile_name=profile_name)
        workdir = pathlib.Path(file).parent.resolve()

        # doing this import here speeds up the CLI app considerably
        from wanna.core.services.jobs import JobService

        job_service = JobService(config=config, workdir=workdir)
        job_service.report(
            instance_name=instance_name,
            wanna_project=config.wanna_project.name,
            wanna_resource="job",
            gcp_project=config.gcp_profile.project_id,
            billing_id=config.wanna_project.billing_id,
            organization_id=config.wanna_project.organization_id,
        )

    @staticmethod
    def _extract_job_overrides(
        extra_args: list[str],
    ) -> tuple[list[str | float | int], list[str]]:
        args: list[str | float | int] = []
        command = []
        for extra_arg in extra_args:
            if extra_arg.startswith("--") or args:
                args.append(extra_arg)
            else:
                command.append(extra_arg)

        return args, command



from pydantic import BaseModel, field_validator

from wanna.core.utils import validators


class CloudSchedulerModel(BaseModel):
    cron: str
    timezone: str = "Etc/UTC"
    service_account: str | None = None

    # Validators
    _schedule = field_validator("cron")(validators.validate_cron_schedule)


class EnvCloudSchedulerModel(CloudSchedulerModel):
    environment: str | None = None



from typing import Annotated

from pydantic import BaseModel, BeforeValidator, Field

from wanna.core.models.docker import DockerModel
from wanna.core.models.gcp_profile import GCPProfileModel
from wanna.core.models.notification_channel import NotificationChannelModel
from wanna.core.models.pipeline import PipelineModel
from wanna.core.models.tensorboard import TensorboardModel
from wanna.core.models.training_custom_job import CustomJobModel, TrainingCustomJobModel
from wanna.core.models.wanna_project import WannaProjectModel
from wanna.core.models.workbench import InstanceModel
from wanna.core.utils.config_enricher import (
    enrich_instance_with_gcp_settings_v2,
)
from wanna.core.utils.validators import validate_docker_images_defined


class WannaConfigModel(BaseModel, validate_assignment=True):
    wanna_project: WannaProjectModel
    gcp_profile: GCPProfileModel
    docker: DockerModel | None = None
    tensorboards: list[
        Annotated[TensorboardModel, BeforeValidator(enrich_instance_with_gcp_settings_v2)]
    ] = Field(default_factory=list)
    jobs: list[
        Annotated[
            CustomJobModel | TrainingCustomJobModel,
            BeforeValidator(enrich_instance_with_gcp_settings_v2),
        ]
    ] = Field(default_factory=list)
    pipelines: list[
        Annotated[PipelineModel, BeforeValidator(enrich_instance_with_gcp_settings_v2)]
    ] = Field(default_factory=list)
    notebooks: list[
        Annotated[
            InstanceModel,
            BeforeValidator(validate_docker_images_defined),
            BeforeValidator(enrich_instance_with_gcp_settings_v2),
        ]
    ] = Field(default_factory=list)
    notification_channels: list[NotificationChannelModel] = Field(default_factory=list)



import pathlib
from pathlib import Path

import typer

from wanna.cli.plugins.base_plugin import BasePlugin
from wanna.cli.plugins.common_options import (
    instance_name_option,
    profile_name_option,
    push_mode_option,
    version_option,
    wanna_file_option,
)
from wanna.core.deployment.models import PushMode
from wanna.core.utils.config_loader import load_config_from_yaml


class PipelinePlugin(BasePlugin):
    """
    Plugin for building and deploying Vertex-AI ML Pipelines.
    """

    def __init__(self) -> None:
        super().__init__()
        self.register_many(
            [
                self.build,
                self.push,
                self.deploy,
                self.run,
                self.run_manifest,
                self.report,
            ]
        )

    @staticmethod
    def build(
        version: str = version_option(instance_type="pipeline"),
        params: Path = typer.Option(
            None,
            "--params",
            envvar="WANNA_ENV_PIPELINE_PARAMS",
            help="Path to the params file in yaml format",
        ),
        file: Path = wanna_file_option,
        profile_name: str = profile_name_option,
        instance_name: str = instance_name_option("pipeline", "compile"),
        mode: PushMode = push_mode_option,
    ) -> None:
        """
        Create a manifest based on the wanna-ml config that can be later pushed, deployed or run.
        """
        config = load_config_from_yaml(file, gcp_profile_name=profile_name)
        workdir = pathlib.Path(file).parent

        # doing this import here speeds up the CLI app considerably
        from wanna.core.services.pipeline import PipelineService

        pipeline_service = PipelineService(
            config=config, workdir=workdir, version=version, push_mode=mode
        )
        pipeline_service.build(instance_name, params)

    @staticmethod
    def push(
        version: str = version_option(instance_type="pipeline"),
        params: Path = typer.Option(
            None,
            "--params",
            envvar="WANNA_ENV_PIPELINE_PARAMS",
            help="Path to the params file in yaml format",
        ),
        file: Path = wanna_file_option,
        profile_name: str = profile_name_option,
        instance_name: str = instance_name_option("pipeline", "push"),
        mode: PushMode = push_mode_option,
    ) -> None:
        """
        Build and push manifest to Cloud Storage.
        """
        config = load_config_from_yaml(file, gcp_profile_name=profile_name)
        workdir = pathlib.Path(file).parent

        # doing this import here speeds up the CLI app considerably
        from wanna.core.services.pipeline import PipelineService

        pipeline_service = PipelineService(
            config=config, workdir=workdir, version=version, push_mode=mode
        )
        manifests = pipeline_service.build(instance_name, params)
        pipeline_service.push(manifests)

    @staticmethod
    def deploy(
        version: str = version_option(instance_type="pipeline"),
        env: str = typer.Option("local", "--env", "-e", envvar="WANNA_ENV", help="Pipeline env"),
        file: Path = wanna_file_option,
        profile_name: str = profile_name_option,
        instance_name: str = instance_name_option("pipeline", "deploy"),
    ) -> None:
        """
        Deploy the pipeline. Deploying means you can set a schedule and the pipeline will not be run only once,
        but on regular basis.
        """
        config = load_config_from_yaml(file, gcp_profile_name=profile_name)
        workdir = pathlib.Path(file).parent

        # doing this import here speeds up the CLI app considerably
        from wanna.core.services.pipeline import PipelineService

        pipeline_service = PipelineService(config=config, workdir=workdir, version=version)
        pipeline_service.deploy(instance_name, env)

    @staticmethod
    def run(
        version: str = version_option(instance_type="pipeline notebook"),
        params: Path = typer.Option(
            None,
            "--params",
            envvar="WANNA_ENV_PIPELINE_PARAMS",
            help="Path to the params file in yaml format",
        ),
        sync: bool = typer.Option(False, "--sync", "-s", help="Runs the pipeline in sync mode"),
        file: Path = wanna_file_option,
        profile_name: str = profile_name_option,
        instance_name: str = instance_name_option("pipeline", "run"),
        mode: PushMode = push_mode_option,
        skip_execution_cache: bool = typer.Option(
            False,
            "--skip-execution-cache",
            help="Kubeflow pipeline cache configuration",
        ),
    ) -> None:
        """
        Run the pipeline as specified in wanna-ml config. This command puts together build, push and run-manifest steps.
        """
        config = load_config_from_yaml(file, gcp_profile_name=profile_name)
        workdir = pathlib.Path(file).parent

        # doing this import here speeds up the CLI app considerably
        from wanna.core.services.pipeline import PipelineService

        pipeline_service = PipelineService(
            config=config,
            workdir=workdir,
            version=version,
            push_mode=mode,
            kubeflow_pipeline_caching=skip_execution_cache,
        )
        manifests = pipeline_service.build(instance_name)
        pipeline_service.push(manifests, local=False)
        PipelineService.run([str(p) for p in manifests], extra_params=params, sync=sync)

    @staticmethod
    def run_manifest(
        manifest: str = typer.Option(None, "--manifest", "-v", help="Job deployment manifest"),
        params: Path = typer.Option(
            None,
            "--params",
            envvar="WANNA_ENV_PIPELINE_PARAMS",
            help="Path to the params file in yaml format",
        ),
        sync: bool = typer.Option(False, "--sync", "-s", help="Runs the pipeline in sync mode"),
    ) -> None:
        """
        Run the pipeline as specified in the wanna-ml manifest.
        """

        # doing this import here speeds up the CLI app considerably
        from wanna.core.services.pipeline import PipelineService

        PipelineService.run([manifest], extra_params=params, sync=sync)

    @staticmethod
    def report(
        file: Path = wanna_file_option,
        profile_name: str = profile_name_option,
        instance_name: str = instance_name_option("pipeline", "report"),
    ) -> None:
        """
        Displays a link to the cost report per wanna_project and optionally per instance name.
        """
        config = load_config_from_yaml(file, gcp_profile_name=profile_name)
        workdir = pathlib.Path(file).parent

        # doing this import here speeds up the CLI app considerably
        from wanna.core.services.pipeline import PipelineService

        pipeline_service = PipelineService(config=config, workdir=workdir)
        pipeline_service.report(
            instance_name=instance_name,
            wanna_project=config.wanna_project.name,
            wanna_resource="pipeline",
            gcp_project=config.gcp_profile.project_id,
            billing_id=config.wanna_project.billing_id,
            organization_id=config.wanna_project.organization_id,
        )



import abc
from typing import Any, Callable, Union

import typer

CommandRegistration = Union[Callable[..., None], tuple[Callable[..., None], dict[Any, Any]]]


class BasePlugin(abc.ABC):
    """Some doc"""

    def __init__(self) -> None:
        self.app = typer.Typer()

    def register_command(self, func: CommandRegistration) -> None:
        if isinstance(func, tuple):
            self.app.command(context_settings=func[1])(func[0])
        else:
            self.app.command()(func)

    def register_many(self, funcs: list[CommandRegistration]) -> None:
        for func in funcs:
            self.register_command(func)



from typing import Any

from pydantic import BaseModel, ConfigDict, EmailStr, field_validator

from wanna.core.utils import validators


class BaseInstanceModel(BaseModel, validate_assignment=True):
    name: str
    project_id: str
    zone: str | None = None
    region: str | None = None
    labels: dict[str, str] | None = None
    description: str | None = None
    service_account: EmailStr | None = None
    network: str | None = None
    bucket: str | None = None
    tags: list[str] | None = None
    metadata: dict[str, Any] | None = None

    model_config = ConfigDict(extra="ignore", validate_assignment=True)

    _project_id = field_validator("project_id", mode="before")(validators.validate_project_id)
    _zone = field_validator("zone", mode="before")(validators.validate_zone)
    _region = field_validator("region", mode="before")(validators.validate_region)
    _network = field_validator("network", mode="before")(validators.validate_network_name)
    _labels = field_validator("labels", mode="before")(validators.validate_labels)



import os
import shutil
from pathlib import Path

from caseconverter import kebabcase
from dirhash import dirhash
from google.api_core.client_options import ClientOptions
from google.api_core.future.polling import DEFAULT_POLLING
from google.cloud.devtools import cloudbuild_v1
from google.cloud.devtools.cloudbuild_v1.services.cloud_build import CloudBuildClient
from google.cloud.devtools.cloudbuild_v1.types import (
    Build,
    BuildOptions,
    BuildStep,
    Source,
    StorageSource,
)
from google.cloud.storage import Blob
from google.protobuf.duration_pb2 import Duration  # pylint: disable=no-name-in-module
from python_on_whales import Image, docker

from wanna.core.deployment.models import PushMode
from wanna.core.loggers.wanna_logger import get_logger
from wanna.core.models.docker import (
    DockerBuildConfigModel,
    DockerImageModel,
    DockerModel,
    ImageBuildType,
    LocalBuildImageModel,
    NotebookReadyImageModel,
    ProvidedImageModel,
)
from wanna.core.models.gcp_profile import GCPProfileModel
from wanna.core.utils import loaders
from wanna.core.utils.credentials import get_credentials
from wanna.core.utils.env import cloud_build_access_allowed, gcp_access_allowed
from wanna.core.utils.gcp import (
    convert_project_id_to_project_number,
    upload_file_to_gcs,
)
from wanna.core.utils.io import tar_docker_context
from wanna.core.utils.templates import render_template

logger = get_logger(__name__)


class DockerClientException(Exception):
    pass


class DockerService:
    def __init__(
        self,
        docker_model: DockerModel,
        gcp_profile: GCPProfileModel,
        version: str,
        work_dir: Path,
        wanna_project_name: str,
        quick_mode: bool = False,  # just returns tags but does not build
    ):
        self.docker_model = docker_model
        self.image_models = docker_model.images
        self.image_store: dict[str, tuple[DockerImageModel, Image | None, str]] = {}

        # Artifactory mirrors to different registry/projectid/repository combo
        registry_suffix = os.getenv("WANNA_DOCKER_REGISTRY_SUFFIX")
        self.docker_registry_suffix = f"{registry_suffix}/" if registry_suffix else ""
        self.docker_registry = (
            os.getenv("WANNA_DOCKER_REGISTRY")
            or docker_model.registry
            or gcp_profile.docker_registry
            or f"{gcp_profile.region}-docker.pkg.dev"
        )
        docker_repository = (
            os.getenv("WANNA_DOCKER_REGISTRY_REPOSITORY") or docker_model.repository
        )
        self.docker_repository = (
            docker_repository if docker_repository else gcp_profile.docker_repository
        )
        self.docker_project_id = (
            os.getenv("WANNA_DOCKER_REGISTRY_PROJECT_ID") or gcp_profile.project_id
        )

        self.version = version
        self.work_dir = work_dir
        self.build_dir = self.work_dir / "build" / "docker"
        self.wanna_project_name = wanna_project_name
        self.project_id = gcp_profile.project_id
        self.location = gcp_profile.region
        self.docker_build_config_path = os.getenv(
            "WANNA_DOCKER_BUILD_CONFIG", self.work_dir / "dockerbuild.yaml"
        )
        self.build_config = self._read_build_config(self.docker_build_config_path)
        self.cloud_build_timeout = docker_model.cloud_build_timeout
        self.cloud_build = (
            False
            if not gcp_access_allowed or not cloud_build_access_allowed
            else docker_model.cloud_build
        )

        self.cloud_build_workerpool = docker_model.cloud_build_workerpool
        self.cloud_build_workerpool_location = (
            docker_model.cloud_build_workerpool_location or self.location
        )
        self.bucket = gcp_profile.bucket
        self.quick_mode = quick_mode
        assert self.cloud_build or self._is_docker_client_active(), DockerClientException(
            "You need running docker client on your machine to use WANNA cli with local docker build"
        )

    def _read_build_config(self, config_path: Path | str) -> DockerBuildConfigModel | None:
        """
        Reads the DockerBuildConfig from local file.
        If the file does not exist, return None.

        Args:
            config_path:

        Returns:

        """
        if os.path.isfile(config_path):
            with open(config_path, encoding="utf-8") as file:
                # Load workflow file
                build_config_dict = loaders.load_yaml(file, self.work_dir)
            build_config = DockerBuildConfigModel.model_validate(build_config_dict)
            return build_config
        return None

    @staticmethod
    def _is_docker_client_active() -> bool:
        """
        Method to find out if you have a running docker client on your machine.

        Returns:
            True if docker client found, False otherwise
        """
        return docker.info().id is not None

    @staticmethod
    def _get_ignore_patterns(context_dir: Path) -> list[str]:
        """
        Get the ignore patterns from .dockerignore file in the context_dir.

        Args:
            context_dir: Path to the docker context directory where .dockerignore file is located

        Returns:
            list of ignore patterns

        """
        docker_ignore = context_dir / ".dockerignore"
        ignores = []

        if docker_ignore.exists():
            with open(docker_ignore, encoding="utf-8") as f:
                lines = f.readlines()
                ignores += [
                    ignore.rstrip()
                    for ignore in lines
                    if not ignore.startswith("#") and not ignore.strip() == ""
                ]
        return ignores

    def _build_image(
        self,
        context_dir,
        file_path: Path,
        tags: list[str],
        docker_image_ref: str,
        **build_args,
    ) -> Image | None:
        """
        Build a docker image locally or in GCP Cloud Build.

        Args:
            context_dir: Path to the directory with all necessary files for docker image build
            file_path: Path to the Dockerfile
            tags: list of tags for the image
            docker_image_ref: Name of the image
            build_args: Additional build arguments

        Returns:
            Image object if the image was built locally, None otherwise
        """

        ignore_patterns = self._get_ignore_patterns(context_dir)

        # skip builds if we are in quick mode or the checksumdir of docker contex_dir has not changed
        skip_build = self._should_skip_build(
            self.build_dir / docker_image_ref, context_dir, ignore_patterns
        )

        if skip_build:
            logger.user_info(
                text=f"Skipping build for context_dir={context_dir}, dockerfile={file_path} and image {tags[0]}"
            )
            return None

        if self.cloud_build:
            logger.user_info(text=f"Building {docker_image_ref} docker image in Cloud build")
            self._build_image_on_gcp_cloud_build(
                context_dir=context_dir,
                file_path=file_path,
                docker_image_ref=docker_image_ref,
                tags=tags,
                ignore_patterns=ignore_patterns,
            )
            return None
        else:
            logger.user_info(
                text=f"Building {docker_image_ref} docker image locally with {build_args}"
            )
            image = docker.build(context_dir, file=file_path, load=True, tags=tags, **build_args)
            self._write_context_dir_checksum(
                self.build_dir / docker_image_ref, context_dir, ignore_patterns
            )
            return image  # type: ignore

    def _pull_image(self, image_url: str) -> Image | None:
        if self.cloud_build or self.quick_mode:
            # TODO: verify that images exists remotely but dont pull them to local
            return None
        else:
            with logger.user_spinner("Pulling image locally"):
                image = docker.pull(image_url, quiet=True)
            return image  # type: ignore

    def find_image_model_by_name(self, image_name: str) -> DockerImageModel:
        """
        Finds a DockerImageModel with given image_name from self.image_models
        Args:
            image_name: name to find

        Returns:
            DockerImageModel
        """
        matched_image_models = list(
            filter(lambda i: i.name.strip() == image_name.strip(), self.image_models)
        )
        if len(matched_image_models) == 0:
            raise ValueError(f"No docker image with name {image_name} found")
        elif len(matched_image_models) > 1:
            raise ValueError(
                f"Multiple docker images with name {image_name} found, please use unique names"
            )
        else:
            return matched_image_models[0]

    def get_image(
        self,
        docker_image_ref: str,
    ) -> tuple[DockerImageModel, Image | None, str]:
        """
        A wrapper around _get_image that checks if the docker image has been already build / pulled

        Args:
            docker_image_ref: Name of the image to get

        Returns:
            tuple of DockerImageModel, Image and image tag

        """

        image = self.image_store.get(docker_image_ref)
        if not image:
            image = self._get_image(docker_image_ref=docker_image_ref)
            self.image_store.update({docker_image_ref: image})

        return image

    def _get_image(
        self,
        docker_image_ref: str,
    ) -> tuple[DockerImageModel, Image | None, str]:
        """
        Given the docker_image_ref, this function prepares the image for you.
        Depending on the build_type, it either build the docker image or
        if you work with provided_image type, it will pull the image to verify the url.

        Args:
            docker_image_ref: Name of the image to get

        Returns:

        """
        docker_image_model = self.find_image_model_by_name(docker_image_ref)

        build_dir = self.work_dir / Path("build") / "docker" / docker_image_model.name
        os.makedirs(build_dir, exist_ok=True)
        build_args = self.build_config.dict() if self.build_config else {}

        if isinstance(docker_image_model, NotebookReadyImageModel):
            tags = self.construct_image_tag(
                image_name=docker_image_model.name,
            )
            template_path = Path("notebook_template.Dockerfile")
            shutil.copy2(
                self.work_dir / docker_image_model.requirements_txt,
                build_dir / "requirements.txt",
            )
            file_path = self._jinja_render_dockerfile(
                docker_image_model, template_path, build_dir=build_dir
            )
            context_dir = build_dir
            image = self._build_image(
                context_dir,
                file_path=file_path,
                tags=tags,
                docker_image_ref=docker_image_ref,
                **build_args,
            )
        elif isinstance(docker_image_model, LocalBuildImageModel):
            tags = self.construct_image_tag(
                image_name=docker_image_model.name,
            )
            file_path = self.work_dir / docker_image_model.dockerfile
            context_dir = self.work_dir / docker_image_model.context_dir
            image = self._build_image(
                context_dir,
                file_path=file_path,
                tags=tags,
                docker_image_ref=docker_image_ref,
                **build_args,
            )
        elif isinstance(docker_image_model, ProvidedImageModel):
            image = self._pull_image(docker_image_model.image_url)
            tags = [docker_image_model.image_url]
        else:
            raise Exception("Invalid image model type.")

        return (
            docker_image_model,
            image,
            tags[0],
        )

    @staticmethod
    def _get_dirhash(directory: Path, ignore_patterns: list[str] | None = None) -> str:
        """
        Get the checksum of the directory.

        Args:
            directory: Path to the directory to be checksummed
            ignore_patterns: list of patterns to ignore

        Returns:
            Checksum of the directory
        """
        return dirhash(directory, "sha256", ignore=set(ignore_patterns or []))

    def _get_cache_path(self, hash_cache_dir: Path) -> Path:
        """
        Get the path to the cache file.

        Args:
            hash_cache_dir: Path to the directory where the cache file is stored
        Returns:
            Path to the cache file
        """
        version = kebabcase(self.version)
        os.makedirs(hash_cache_dir, exist_ok=True)
        return hash_cache_dir / f"{kebabcase(self.docker_repository)}-{version}-cache.sha256"

    def _should_skip_build(
        self, hash_cache_dir: Path, context_dir: Path, ignore_patterns: list[str]
    ) -> bool:
        """
        Check if the context_dir has changed since the last build or if quick mode is enabled
        to decide if wanna should build the image again.

        Args:
            hash_cache_dir: Path to the directory where the cache file is stored
            context_dir: Path to the directory to check
            ignore_patterns: list of patterns to ignore

        Returns:
            True if the context_dir has changed, False otherwise
        """
        cache_file = self._get_cache_path(hash_cache_dir)
        sha256hash = self._get_dirhash(context_dir, ignore_patterns)
        context_dir_hash_match = False

        if cache_file.exists():
            with open(cache_file, encoding="utf-8") as f:
                old_hash = f.read().replace("\n", "")
                context_dir_hash_match = old_hash == sha256hash

        return self.quick_mode or context_dir_hash_match

    def _write_context_dir_checksum(
        self,
        hash_cache_dir: Path,
        context_dir: Path,
        ignore_patterns: list[str] | None = None,
    ) -> None:
        """
        Write the checksum of the context_dir to a file.

        Args:
            hash_cache_dir: Path to the directory where the cache file is stored
            context_dir: Path to the directory to check
            ignore_patterns: list of patterns to ignore
        Returns:
            None
        """

        cache_file = self._get_cache_path(hash_cache_dir)
        sha256hash = self._get_dirhash(context_dir, ignore_patterns)
        with open(cache_file, "w", encoding="utf-8") as f:
            f.write(sha256hash)

    def _build_image_on_gcp_cloud_build(
        self,
        context_dir: Path,
        file_path: Path,
        tags: list[str],
        docker_image_ref: str,
        ignore_patterns: list[str] | None = None,
    ) -> None:
        """
        Build a docker container in GCP Cloud Build and push the images to registry.
        Folder context_dir is tarred, uploaded to GCS and then used to building.

        Args:
            context_dir: directory with all necessary files for docker image build
            file_path: path to Dockerfile
            tags: list of tags for the image
            docker_image_ref: Name of the image
            ignore_patterns: list of patterns to ignore

        Returns:
            None

        """

        # Set the pooling timeout to self.cloud_build_timeout seconds
        # since often large GPUs builds exceed the 900s limit
        timeout = Duration()
        timeout.seconds = self.cloud_build_timeout
        DEFAULT_POLLING._timeout = self.cloud_build_timeout
        project_number = convert_project_id_to_project_number(self.project_id)

        dockerfile = os.path.relpath(file_path, context_dir)
        blob = self._upload_context_dir_to_gcs(context_dir, docker_image_ref, ignore_patterns)

        tags_args = " ".join([f"--destination={t}" for t in tags]).split()
        kaniko_build_args = (
            tags_args + self.docker_model.cloud_build_kaniko_flags + ["--dockerfile", dockerfile]
        )

        options, api_endpoint = (
            (None, "cloudbuild.googleapis.com")
            if not self.cloud_build_workerpool
            else (
                BuildOptions(
                    pool=BuildOptions.PoolOption(
                        name=f"projects/{project_number}/locations/{self.cloud_build_workerpool_location}/workerPools/{self.cloud_build_workerpool}"
                    )
                ),
                f"{self.cloud_build_workerpool_location}-cloudbuild.googleapis.com",
            )
        )

        build = Build(
            source=Source(
                storage_source=StorageSource(bucket=blob.bucket.name, object_=blob.name)
            ),
            steps=[
                BuildStep(
                    name=f"gcr.io/kaniko-project/executor:{self.docker_model.cloud_build_kaniko_version}",
                    args=kaniko_build_args,
                )
            ],
            timeout=timeout,
            options=options,
        )
        client = CloudBuildClient(
            credentials=get_credentials(),
            client_options=ClientOptions(api_endpoint=api_endpoint),
        )
        request = cloudbuild_v1.CreateBuildRequest(project_id=self.project_id, build=build)

        op = client.create_build(request=request)
        build_id = op.metadata.build.id
        base = "https://console.cloud.google.com/cloud-build/builds"
        link = base + (
            f";region={self.cloud_build_workerpool_location}/{build_id}?project={project_number}"
            if self.cloud_build_workerpool
            else f"/{build_id}?project={project_number}"
        )

        logger.user_info(text=f"Build started {link}")

        try:
            op.result()
            self._write_context_dir_checksum(
                self.build_dir / docker_image_ref, context_dir, ignore_patterns
            )
        except:
            raise Exception(f"Build failed {link}")

    def push_image(self, image_or_tags: Image | list[str], quiet: bool = False) -> None:
        """
        Push a docker image to the registry (image must have tags)
        If you are in the cloud_build mode, nothing is pushed, images already live in cloud.

        Args:
            image_or_tags: what docker resource to push
            quiet: If you don't want to see the progress bars.

        Returns:
            None
        """

        if not self.cloud_build:
            tags = image_or_tags.repo_tags if isinstance(image_or_tags, Image) else image_or_tags
            logger.user_info(text=f"Pushing docker image {tags}")
            docker.image.push(tags, quiet)

    def push_image_ref(
        self,
        image_ref: str,
        quiet: bool = False,  # noqa: ARG002
    ) -> None:
        """
        Push a docker image ref to the registry (image must have tags)

        Args:
            image_ref: image_ref to push
            quiet: If you don't want to see the progress bars.

        Returns:
            None
        """
        model, image, tag = self.get_image(image_ref)
        if (image or tag) and model.build_type != ImageBuildType.provided_image:
            self.push_image(image or [tag])

    @staticmethod
    def remove_image(image: Image, force=False, prune=True) -> None:
        """
        Remove docker image, useful if you don't want to clutter your machine.

        Args:
            image: Image to remove
            force: Force remove
            prune: Prune

        Returns:
            None

        """
        docker.image.remove(image, force=force, prune=prune)

    def construct_image_tag(
        self,
        image_name: str,
    ):
        """
        Construct full image tag.

        Args:
            image_name: name of the image

        Returns:
            list of full image tag
        """
        versions = [self.version, "latest"]
        return [
            f"{self.docker_registry}/{self.docker_registry_suffix}{self.docker_project_id}/"
            f"{self.docker_repository}/{self.wanna_project_name}/{image_name}:{version}"
            for version in versions
        ]

    def build_container_and_get_image_url(
        self, docker_image_ref: str, push_mode: PushMode = PushMode.all
    ) -> str:
        """
        Build a docker image and push it to the registry.

        :param docker_image_ref:
        :param push_mode:
        :return:
        """
        if push_mode == PushMode.quick:
            # only get image tag
            docker_image_model = self.find_image_model_by_name(docker_image_ref)
            tags = self.construct_image_tag(image_name=docker_image_model.name)
            image_url = tags[0]
        elif push_mode == PushMode.manifests:
            # only build image
            image_tag = self.get_image(docker_image_ref=docker_image_ref)
            image_url = image_tag[2]
        else:
            # build image and push
            image_tag = self.get_image(docker_image_ref=docker_image_ref)
            if len(image_tag) > 1 and image_tag[1]:
                self.push_image(image_tag[1])
            image_url = image_tag[2]
        return image_url

    @staticmethod
    def _jinja_render_dockerfile(
        image_model: DockerImageModel,
        template_path: Path,
        build_dir: Path,
    ) -> Path:
        """
        Based on image_model.type, we render dockerfile for this image type.
        Args:
            image_model: docker image model
            template_path: path to the dockerfile jinja template
            build_dir: build directory (where to save rendered dockerfile)

        Returns:
            path to the rendered dockerfile
        """
        if isinstance(image_model, NotebookReadyImageModel):
            rendered = render_template(
                source_path=template_path,
                base_image=image_model.base_image,
                requirements_txt=image_model.requirements_txt,
            )
        else:
            raise Exception("Invalid docker image type.")

        docker_file_path = build_dir / Path(f"{image_model.name}.Dockerfile")

        with open(docker_file_path, "w", encoding="utf-8") as file:
            file.write(rendered)

        return docker_file_path

    def _upload_context_dir_to_gcs(
        self,
        context_dir: Path,
        docker_image_ref: str,
        ignore_patterns: list[str] | None = None,
    ) -> Blob:
        """
        Tar the context_dir and upload it to GCS.

        Args:
            context_dir: Path to the directory to be tarred and uploaded
            docker_image_ref: Name of the image
            ignore_patterns: list of patterns to ignore whilst tarring

        Returns:
            Blob
        """
        tar_filename = self.work_dir / "build" / "docker" / f"{docker_image_ref}.tar.gz"
        tar_docker_context(context_dir, tar_filename, ignore_patterns or [])

        blob_name = os.path.relpath(tar_filename, self.work_dir).replace("\\", "/")
        blob = upload_file_to_gcs(
            filename=tar_filename, bucket_name=self.bucket, blob_name=blob_name
        )
        return blob



from pydantic import BaseModel, ConfigDict, EmailStr


class WannaProjectModel(BaseModel):
    """
    `wanna_project` section of the yaml config consists of the following inputs:

    - `name` - [str] the name of the wanna project should be unique, this name will be used in the docker service
      for naming docker images and in labeling GCP resources. Hence it can be used also for budget monitoring.
    - `version` - [str] Currently used only in labeling GCP resources, we expect to introduce new API versions
      and then this parameter will gain more importance.
    - `authors` - [list[str]] Email addresses, currently used only in GCP resource labeling but soon also in monitoring.
    - `billing_id` - [str] (optional) GCP Billing ID, needed for the budget report
    - `organization_id` - [str] (optional) GCP Organization ID, needed for the budget report
    """

    name: str
    version: str
    authors: list[EmailStr]
    billing_id: str | None = None
    organization_id: str | None = None

    model_config = ConfigDict(extra="forbid", coerce_numbers_to_str=True)



from typing import Any, cast

from pydantic_core.core_schema import ValidationInfo

from wanna.core.models.gcp_profile import GCPProfileModel
from wanna.core.models.wanna_project import WannaProjectModel
from wanna.core.utils.credentials import get_gcloud_user


def add_labels(instance_dict: dict[str, Any], new_labels: dict[str, str]) -> dict[str, Any]:
    """
    Add new labels to the instance model.
    Args:
        instance_dict: dictionary representing one instance
        new_labels: new labels to be added

    Returns:
        instance_dict: dictionary representing one instance with added labels
    """
    labels = instance_dict.get("labels") or {}
    labels.update(new_labels)
    instance_dict["labels"] = labels
    return instance_dict


def email_fixer(email: str) -> str:
    return email.replace(".", "-").replace("@", "_at_")


def generate_default_labels(wanna_project: WannaProjectModel) -> dict[str, str]:
    """
    Get the default labels (GCP labels) that will be used with all instances based on wanna_project info.
    Args:
        wanna_project

    Returns:
        default labels
    """
    return {
        "wanna_project": wanna_project.name,
        "wanna_project_version": str(wanna_project.version).replace(".", "__"),
        "wanna_project_authors": "_".join(
            [email_fixer(author.partition("@")[0]) for author in wanna_project.authors]
        ),
        "author": email_fixer(get_gcloud_user()),
    }


def enrich_instance_info_with_gcp_settings_dict(
    instance_dict: dict[str, Any], gcp_profile: GCPProfileModel
) -> dict[str, Any]:
    """
    The dictionary instance_dict is updated with values from gcp_settings. This allows you to set values such as
    project_id and zone only on the wanna-ml config level but also give you the freedom to set separately for each
    notebook, jobs, etc. The values as at the instance level take precedence over general wanna-ml settings.

    Args:
        instance_dict: dict with values from wanna-ml config from one instance (one job, one notebook)
        gcp_profile: GCPSettings model

    Returns:
        dict: enriched with general gcp_settings if those information was not set on instance level

    """
    gcp_settings_dict = gcp_profile.model_dump().copy()
    gcp_settings_dict = {k: v for k, v in gcp_settings_dict.items() if v is not None}
    instance_info = gcp_settings_dict
    instance_info.update(instance_dict)
    return instance_info


def enrich_gcp_profile_with_wanna_default_labels(
    cls,
    values_inst,
    values,  # noqa: ARG001
):
    """
    Enrich gcp_profile with wanna default project labels

    Args:
        cls: pydantic class
        values_inst: values representing gcp_profile
        values: values for the whole pydantic object, loaded so far
    Returns:
        values_inst: updated gcp_profile
    """
    labels = generate_default_labels(wanna_project=values.get("wanna_project"))
    values_inst.labels = {**values_inst.labels, **labels}
    return values_inst


def enrich_instance_with_gcp_settings_v2(value, info: ValidationInfo):  # noqa: ARG001
    """
    Enrich dictionary representing one instance (one notebook, job, etc.)
    with information wanna_project and gcp_settings. This is useful in scenario when
    some parameters (project_id) are needed for all instances, but you can set it
    only on general gcp_settings level.

    Args:
        cls: pydantic class
        values_inst: values representing one instance
        values: values for the whole pydantic object

    Returns:
        values_inst: values representing one instance enriched with information from wanna_project and gcp_settings
    """
    value = enrich_instance_info_with_gcp_settings_dict(
        instance_dict=value, gcp_profile=cast(GCPProfileModel, info.data.get("gcp_profile"))
    )
    labels = generate_default_labels(
        wanna_project=cast(WannaProjectModel, info.data.get("wanna_project"))
    )
    value = add_labels(instance_dict=value, new_labels=labels)
    return value



import os
from pathlib import Path
from string import Template

import kfp.components as comp


def load_wanna_component(path: Path | str):
    with open(str(path), encoding="utf-8") as f:
        t = Template(f.read())
        component = t.safe_substitute(os.environ)
        return comp.load_component_from_text(component)



from abc import abstractmethod
from typing import TypeVar, Union

import typer
from google.api_core import exceptions
from google.api_core.operation import Operation
from google.cloud.notebooks_v1 import CreateInstanceRequest as CreateInstanceRequestV1
from google.cloud.notebooks_v1 import CreateRuntimeRequest, Runtime
from google.cloud.notebooks_v1 import Instance as InstanceV1
from google.cloud.notebooks_v2 import CreateInstanceRequest, Instance
from waiting import wait

from wanna.core.deployment.models import PushMode
from wanna.core.loggers.wanna_logger import get_logger
from wanna.core.models.workbench import BaseWorkbenchModel
from wanna.core.services.base import BaseService

T = TypeVar("T", bound=BaseWorkbenchModel)
CreateRequest = Union[CreateRuntimeRequest, CreateInstanceRequestV1, CreateInstanceRequest]
Instances = Union[Runtime, InstanceV1, Instance]

logger = get_logger(__name__)


class BaseWorkbenchService(BaseService[T]):
    @abstractmethod
    def _instance_exists(self, instance: T) -> bool:
        """
        Check if the instance with given instance_name exists in given GCP project project_id and location.
        Args:
            instance: notebook to verify if exists on GCP

        Returns:
            True if exists, False if not
        """

    def _create_one_instance(self, instance: T, **kwargs) -> None:
        """
        Create a notebook instance based on information in NotebookModel class.
        1. Check if the notebook already exists
        2. Parse the information from NotebookModel to GCP API friendly format = runtime_request
        3. Wait for the compute instance behind the notebook to start
        4. Wait for JupyterLab to start
        5. Get and print the link to JupyterLab

        Args:
            instance: notebook to be created

        """
        should_end = self._create_one_instance_exists(instance)
        if should_end:
            return
        push_mode: PushMode = kwargs.get("push_mode")  # type: ignore
        request = self._create_instance_request(
            instance=instance, deploy=True, push_mode=push_mode
        )
        logger.user_info(f"Creating underlying compute engine instance for {instance.name} ...")
        nb_instance = self._create_instance_client(request=request)
        instance_full_name = (
            nb_instance.result().name
        )  # .result() waits for compute engine behind the notebook to start
        logger.user_info(f"Starting JupyterLab for {instance.name} ...")
        wait(
            lambda: self._validate_jupyterlab_state(instance_full_name, Runtime.State.ACTIVE),
            timeout_seconds=450,
            sleep_seconds=20,
            waiting_for="Starting JupyterLab in your instance",
        )
        jupyterlab_link = self._get_jupyterlab_link(instance_full_name)
        logger.user_success(f"JupyterLab for {instance.name} started at {jupyterlab_link}")

    def _create_one_instance_exists(self, instance: T) -> bool:
        """
        Checks if the instance already exists and prompts the user to delete it if it does.

        Args:
            instance: Workbench instance to check.

        Returns:
            should_end: True if the method should end.
        """
        exists = self._instance_exists(instance)
        if exists:
            logger.user_info(
                f"{self.instance_type} {instance.name} already exists in location {self.workbench_location(instance)}"
            )
            should_recreate = typer.confirm("Are you sure you want to delete it and start a new?")
            if should_recreate:
                self._delete_one_instance(instance)
            else:
                return True
        return False

    @abstractmethod
    def workbench_location(self, instance: T) -> str:
        """
        Args:
            instance: workbench instance.

        Returns:
            location: Return the location of the workbench instance, either region or zone.
        """

    @abstractmethod
    def _create_instance_request(
        self,
        instance: T,
        deploy: bool = True,
        push_mode: PushMode = PushMode.all,
    ) -> CreateRequest:
        """
        Transform the information about desired notebook from our NotebookModel model (based on yaml config)
        to the form suitable for GCP API.

        Args:
            instance: Workbench instance to create.
            deploy: If True, deploy the instance.
            push_mode: Push mode to use.

        Returns:
            request: Return the instance creation request.
        """

    def _delete_one_instance(self, instance: T) -> None:
        """
        Delete one notebook instance. This assumes that it has been already verified that notebook exists.

        Args:
            instance: notebook to delete
        """

        exists = self._instance_exists(instance)
        if exists:
            logger.user_info(f"Deleting {self.instance_type} {instance.name} ...")
            deleted = self._delete_instance_client(instance=instance)
            deleted.result()
            logger.user_success(f"Deleted {self.instance_type} {instance.name}")
        else:
            logger.user_error(
                f"{self.instance_type} with name {instance.name} was not found in location {self.workbench_location(instance)}",
            )
            typer.Exit(1)

    @abstractmethod
    def _delete_instance_client(self, instance: T) -> Operation:
        """
        Actually delete the instance from Vertex AI.

        Args:
            instance: Workbench instance to delete.

        Returns:
            operation: Operation for the deletion.
        """

    @abstractmethod
    def _create_instance_client(self, request: CreateRequest) -> Operation:
        """
        Actually create the instance in Vertex AI.

        Args:
            request: Request for creation

        Returns:
            operation: Operation for the creation.
        """

    def _validate_jupyterlab_state(self, instance_id: str, state: int) -> bool:
        """
        Validate if the given notebook instance is in given state.

        Args:
            instance_id: Full notebook instance id
            state: Notebook state (ACTIVE, PENDING,...)

        Returns:
            True if desired state, False otherwise
        """
        try:
            instance_info = self._client_get_instance(instance_id)
        except exceptions.NotFound:
            raise exceptions.NotFound(
                f"{self.instance_type} {instance_id} was not found."
            ) from None
        return instance_info.state == state

    @abstractmethod
    def _client_get_instance(self, instance_id: str) -> Instances:
        """
        Gets instance by id.

        Args:
            instance_id:

        Returns:
            instance: the instance
        """

    @abstractmethod
    def _get_jupyterlab_link(self, instance_id: str) -> str:
        """
        Get a link to jupyterlab proxy based on given notebook instance id.
        Args:
            instance_id: full notebook instance id

        Returns:
            proxy_uri: link to jupyterlab
        """



from enum import Enum
from pathlib import Path
from typing import Any, Generic, Literal, TypeVar, Union

from pydantic import BaseModel, ConfigDict, EmailStr, Field

from wanna.core.models.cloud_scheduler import CloudSchedulerModel, EnvCloudSchedulerModel
from wanna.core.models.docker import DockerBuildResult
from wanna.core.models.notification_channel import NotificationChannelModel
from wanna.core.models.training_custom_job import JobModelTypeAlias

PipelineEnvParams = dict[str, Union[str, None, EmailStr]]


class GCPResource(BaseModel):
    name: str
    project: str
    location: str
    service_account: str | None = None

    model_config = ConfigDict(
        extra="forbid", validate_assignment=True, arbitrary_types_allowed=True
    )

    def get_base_resource(self):
        return self.model_dump()


class NotificationChannelResource(GCPResource):
    type_: Literal["email", "pubsub"]
    config: dict[str, str]
    labels: dict[str, str]
    description: str | None = None


class CloudSchedulerResource(GCPResource):
    cloud_scheduler: CloudSchedulerModel
    body: dict[str, Any]
    labels: dict[str, str]
    notification_channels: list[str]


class CloudFunctionResource(GCPResource):
    build_dir: Path
    resource_root: str
    resource_function_template: str
    resource_requirements_template: str
    template_vars: dict[str, Any]
    env_params: PipelineEnvParams
    labels: dict[str, str]
    network: str | None = None
    notification_channels: list[str]


class LogMetricResource(GCPResource):
    filter_: str
    description: str


class AlertPolicyResource(GCPResource):
    logging_metric_type: str
    resource_type: str
    display_name: str
    labels: dict[str, str]
    notification_channels: list[str]


class PipelineResource(GCPResource):
    pipeline_name: str
    pipeline_bucket: str
    pipeline_root: str
    pipeline_version: str
    json_spec_path: str
    parameter_values: dict[str, Any] = Field(default_factory=dict)
    labels: dict[str, str] = Field(default_factory=dict)
    enable_caching: bool = True
    schedule: CloudSchedulerModel | list[EnvCloudSchedulerModel] | None
    docker_refs: list[DockerBuildResult]
    compile_env_params: PipelineEnvParams
    network: str | None = None
    notification_channels: list[NotificationChannelModel] = Field(default_factory=list)
    encryption_spec_key_name: str | None = None
    experiment: str | None = None


# BaseCustomJobModel
JOB = TypeVar("JOB", bound=JobModelTypeAlias)  # dependency from wanna models


class JobResource(GCPResource, Generic[JOB]):
    job_payload: dict[str, Any]
    image_refs: list[str] = Field(default_factory=list)
    tensorboard: str | None = None
    network: str | None = None
    job_config: JOB
    encryption_spec: str | None = None
    environment_variables: dict[str, str] | None = None


class PushArtifact(BaseModel):
    name: str


class JsonArtifact(PushArtifact):
    name: str
    json_body: dict[Any, Any]
    destination: str


class PathArtifact(PushArtifact):
    name: str
    source: str
    destination: str


class ContainerArtifact(PushArtifact):
    name: str
    tags: list[str]


class PushMode(str, Enum):
    all = "all"
    manifests = "manifests"
    containers = "containers"
    quick = "quick"

    def is_quick_mode(self) -> bool:
        return self == PushMode.manifests or self == PushMode.quick

    def can_push_containers(self) -> bool:
        return self == PushMode.all or self == PushMode.containers

    def can_push_gcp_resources(self) -> bool:
        return self == PushMode.all or self == PushMode.manifests


class PushTask(BaseModel):
    manifest_artifacts: list[PathArtifact]
    container_artifacts: list[ContainerArtifact]
    json_artifacts: list[JsonArtifact]


PushResult = list[tuple[list[ContainerArtifact], list[PathArtifact], list[JsonArtifact]]]



import logging
import re

from cron_validator import CronValidator
from google.api_core import exceptions
from google.cloud.notebooks_v1.types.instance import Instance
from google.cloud.storage import Client as StorageClient
from pydantic_core.core_schema import ValidationInfo

from wanna.core.models.docker import DockerModel
from wanna.core.utils.credentials import get_credentials
from wanna.core.utils.env import should_validate
from wanna.core.utils.gcp import (
    get_available_compute_machine_types,
    get_available_regions,
    get_available_zones,
    get_network_info,
)


def validate_docker_images_defined(value, info: ValidationInfo):
    docker_image_ref = info.data.get("environment", {}).get("docker_image_ref")
    if docker_image_ref:
        if not info.data.get("docker"):
            raise ValueError(f"Docker image with name {docker_image_ref} is not defined")
        docker_configuration: DockerModel | None = info.data.get("docker")
        defined_images = (
            [i.name for i in docker_configuration.images] if docker_configuration else []
        )
        if docker_image_ref not in defined_images:
            raise ValueError(f"Docker image with name {docker_image_ref} is not defined")
    return value


def validate_zone(zone, values):
    available_zones = get_available_zones(project_id=values.data.get("project_id"))
    if zone not in available_zones:
        raise ValueError(f"Zone invalid ({zone}). must be on of: {available_zones}")
    return zone


def validate_region(region, values):
    available_regions = get_available_regions(project_id=values.data.get("project_id"))
    if region not in available_regions:
        raise ValueError(f"Region invalid ({region}). must be on of: {available_regions}")
    return region


def validate_machine_type(machine_type, values):
    available_machine_types = get_available_compute_machine_types(
        project_id=values.data.get("project_id"), zone=values.data.get("zone")
    )
    if machine_type not in available_machine_types:
        raise ValueError(
            f"Machine type invalid ({machine_type}). must be on of: {available_machine_types}"
        )
    return machine_type


def validate_network_name(network_name: str | None):
    if network_name and not get_network_info(network_name):
        if not re.match("^[a-z][a-z0-9-]+$", network_name):
            raise ValueError(
                "Invalid format of network name. Either use the full name of VPC network"
                "'projects/{project_id}/global/networks/{network_id}'"
                "or just '{network_id}'. In the second case, the project_id will be parsed from notebook settings."
            )
    return network_name


def validate_bucket_name(bucket_name):
    if should_validate:
        try:
            _ = StorageClient(credentials=get_credentials()).get_bucket(bucket_name)
        except exceptions.NotFound:
            raise ValueError(f"Bucket with name {bucket_name} does not exist")
        except exceptions.Forbidden:
            logging.warning(f"Your user does not have permission to access bucket {bucket_name}")

    return bucket_name


def validate_only_one_must_be_set(cls, v):  # noqa: ARG001
    items_set = {key for key, value in v.items() if value is not None}
    if len(items_set) == 0:
        raise ValueError(f"One of {list(v.keys())} must be set.")
    elif len(items_set) > 1:
        raise ValueError(f"Specify only one of {items_set}")
    return v


def validate_cron_schedule(schedule: str):
    if schedule is not None and CronValidator.parse(schedule) is None:
        raise ValueError(f"Cron expression is invalid ({schedule}).")
    else:
        return schedule


def validate_disk_type(disk_type):
    disk_type = disk_type.upper()
    if disk_type not in Instance.DiskType.__members__:
        raise ValueError(
            f"Disk type invalid ({type}). must be on of: {Instance.DiskType._member_names_}"
        )
    return disk_type


def validate_accelerator_type(
    accelerator_type: Instance.AcceleratorType,
) -> Instance.AcceleratorType:
    if accelerator_type not in Instance.AcceleratorType.__members__:
        raise ValueError(
            f"GPU accelerator type invalid ({accelerator_type})."
            f"must be on of: {Instance.AcceleratorType._member_names_}"
        )
    return accelerator_type


def validate_project_id(project_id: str) -> str:
    if not re.match("^[a-z][a-z0-9-]{4,28}[a-z0-9]$", project_id):
        raise ValueError(
            "Invalid GCP project id. project_id: "
            "Must be 6 to 30 characters in length. "
            "Contains lowercase letters, numbers, and hyphens. "
            "Must start with a letter. "
            "Cannot end with a hyphen."
        )
    return project_id


def validate_labels(labels: dict[str, str]) -> dict[str, str]:
    if not labels:
        return {}

    for key, value in labels.items():
        if not re.match(r"^[a-z]{1}[a-z0-9_-]{0,62}$", key) or not re.match(
            r"^[a-z0-9_-]{0,63}$", value
        ):
            raise ValueError(
                "Invalid custom label!"
                "Keys and values can contain only lowercase letters, numeric characters,"
                "underscores, and dashes. Max length is 63 characters."
                "https://cloud.google.com/compute/docs/labeling-resources#requirements"
            )
    return labels



import os
from pathlib import Path

from caseconverter import kebabcase


class JobPaths:
    job_manifest_filename = "job-manifest.json"

    def __init__(self, workdir: Path, bucket: str, job_name: str):
        self.pipeline_name = job_name
        self.workdir = workdir
        self.bucket = bucket
        self.pipelines_dir = (workdir / "build").resolve()
        self.local_job_path = self._get_local_job_path(self.pipelines_dir, job_name)
        self.gcs_job_path = self._get_gcs_job_path(self.bucket, job_name)

    def _get_local_job_path(self, base_path: Path, pipeline_name: str) -> Path:
        os.makedirs(base_path, exist_ok=True)
        return base_path / "wanna-jobs" / kebabcase(pipeline_name).lower()

    def _get_gcs_job_path(self, base_path: str, pipeline_name: str) -> str:
        return f"{base_path}/wanna-jobs/{kebabcase(pipeline_name).lower()}"

    def _get_local_job_manifests_path(self, base_path: Path, version: str) -> Path:
        path = base_path / "deployment" / version / "manifests"
        os.makedirs(path, exist_ok=True)
        return path

    def _get_gcs_job_manifests_path(self, base_path: str, version: str) -> str:
        path = f"{base_path}/deployment/{version}/manifests"
        return path

    def get_local_job_manifest_path(self, version: str) -> str:
        return str(self._get_local_job_manifests_path(self.local_job_path, version))

    def get_gcs_job_manifest_path(self, version: str) -> str:
        return self._get_gcs_job_manifests_path(self.gcs_job_path, version)

    def get_local_job_wanna_manifest_path(self, version: str) -> str:
        path = Path(self.get_local_job_manifest_path(version)) / self.job_manifest_filename
        return str(path)

    def get_gcs_job_wanna_manifest_path(self, version: str) -> str:
        return f"{self.get_gcs_job_manifest_path(version)}/{self.job_manifest_filename}"


class PipelinePaths:
    json_spec_filename = "pipeline-spec.json"
    wanna_manifest_filename = "wanna-manifest.json"
    job_manifest_filename = "job-manifest.json"

    def __init__(self, workdir: Path, bucket: str, pipeline_name: str):
        self.pipeline_name = pipeline_name
        self.workdir = workdir
        self.bucket = bucket
        self.pipelines_dir = (workdir / "build").resolve()
        self.local_pipeline_path = self._get_local_pipeline_path(self.pipelines_dir, pipeline_name)
        self.gcs_pipeline_path = self._get_gcs_pipeline_path(self.bucket, pipeline_name)
        self.local_job_path = self._get_local_pipeline_path(self.pipelines_dir, pipeline_name)
        self.gcs_job_path = self._get_gcs_pipeline_path(self.bucket, pipeline_name)

    def _get_local_pipeline_path(self, base_path: Path, pipeline_name: str) -> Path:
        os.makedirs(base_path, exist_ok=True)
        return base_path / "wanna-pipelines" / kebabcase(pipeline_name).lower()

    def _get_gcs_pipeline_path(self, base_path: str, pipeline_name: str) -> str:
        return f"{base_path}/wanna-pipelines/{kebabcase(pipeline_name).lower()}"

    def get_local_pipeline_manifest_path(self, version: str) -> str:
        path = self.local_pipeline_path / "deployment" / version / "manifests"
        os.makedirs(path, exist_ok=True)
        return str(path)

    def get_gcs_pipeline_manifest_path(self, version: str) -> str:
        path = f"{self.gcs_pipeline_path}/deployment/{version}/manifests"
        return path

    def get_local_pipeline_deployment_path(self, version: str) -> Path:
        path = self.local_pipeline_path / "deployment" / version
        os.makedirs(path, exist_ok=True)
        return path

    def get_gcs_pipeline_deployment_path(self, version: str) -> str:
        path = f"{self.gcs_pipeline_path}/deployment/{version}"
        return path

    def get_local_pipeline_json_spec_path(self, version: str) -> str:
        path = Path(self.get_local_pipeline_manifest_path(version)) / self.json_spec_filename
        return str(path)

    def get_gcs_pipeline_json_spec_path(self, version: str) -> str:
        return f"{self.get_gcs_pipeline_manifest_path(version)}/{self.json_spec_filename}"

    def get_local_wanna_manifest_path(self, version: str) -> str:
        path = Path(self.get_local_pipeline_manifest_path(version)) / self.wanna_manifest_filename
        return str(path)

    def get_gcs_wanna_manifest_path(self, version: str) -> str:
        return f"{self.get_gcs_pipeline_manifest_path(version)}/{self.wanna_manifest_filename}"

    def get_gcs_pipeline_root(self) -> str:
        return f"{self.gcs_pipeline_path}/executions/"

    def get_local_pipeline_root(self) -> str:
        path = self.local_pipeline_path / "executions"
        return str(path)



from typing import Literal

from pydantic import BaseModel, ConfigDict, Field, field_validator

from wanna.core.utils import validators


class VMImage(BaseModel):
    version: str | None = None

    model_config = ConfigDict(extra="forbid")


class Disk(BaseModel):
    disk_type: str
    size_gb: int = Field(default=100, ge=100, le=65535)

    model_config = ConfigDict(extra="forbid")

    _disk_type = field_validator("disk_type")(validators.validate_disk_type)


class GPU(BaseModel):
    count: Literal[1, 2, 4, 8]
    accelerator_type: str
    install_gpu_driver: bool = True
    custom_gpu_driver_path: str | None = None

    model_config = ConfigDict(extra="forbid")

    _accelerator_type = field_validator("accelerator_type")(validators.validate_accelerator_type)



import os
import pathlib
import sys
from io import StringIO
from pathlib import Path
from typing import Any, ParamSpec, TypeVar

from wanna.core.loggers.wanna_logger import get_logger
from wanna.core.models.gcp_profile import GCPProfileModel
from wanna.core.models.wanna_config import WannaConfigModel
from wanna.core.models.wanna_project import WannaProjectModel
from wanna.core.utils import loaders
from wanna.core.utils.env import gcp_access_allowed
from wanna.core.utils.gcp import verify_gcloud_presence

logger = get_logger(__name__)


P = ParamSpec("P")
T = TypeVar("T")


def load_gcp_profile(profile_name: str, wanna_dict: dict[str, Any]) -> GCPProfileModel:
    """
    This functions goes through wanna-ml config and optionally through a file
    $WANNA_GCP_PROFILE_PATH, reads all gcp profiles and returns the one
    with "profile_name" name
    Args:
        profile_name: name of the GCP profile
        wanna_dict: wanna-ml configuration as a dictionary
    Returns:
        GCPProfileModel

    """
    profiles = wanna_dict.get("gcp_profiles", {})
    profiles = {p.get("profile_name"): p for p in profiles}

    extra_profiles_path = os.getenv("WANNA_GCP_PROFILE_PATH")
    if extra_profiles_path and os.path.isfile(extra_profiles_path):
        extra_profiles = loaders.load_yaml_path(
            Path(extra_profiles_path), Path(extra_profiles_path).parent.absolute()
        ).get("gcp_profiles")
        if extra_profiles:
            extra_profiles = {p.get("profile_name"): p for p in extra_profiles}
            profiles.update(extra_profiles)

    if profile_name not in profiles:
        raise ValueError(f"Profile {profile_name} not found")

    selected_profile = profiles.get(profile_name)
    profile_model = GCPProfileModel.model_validate(selected_profile)
    return profile_model


def load_config_from_yaml(
    wanna_config_path: Path | str, gcp_profile_name: str
) -> WannaConfigModel:
    """
    Load the yaml file from wanna_config_path and parses the information to the models.
    This also includes the data validation.

    Args:
        wanna_config_path: path to the wanna-ml yaml file
        gcp_profile_name: name of the GCP profile

    Returns:
        WannaConfigModel

    """
    wanna_config_path = (
        Path(wanna_config_path) if isinstance(wanna_config_path, str) else wanna_config_path
    )
    verify_gcloud_presence()
    with logger.user_spinner("Reading and validating wanna yaml config"):
        wanna_dict = load_yaml_maybe_stdin(wanna_config_path)

        # GCP Profile & Wanna Project metadata is required to be validated first, since are used as enrichers
        profile_model = load_gcp_profile(profile_name=gcp_profile_name, wanna_dict=wanna_dict)
        os.environ["GOOGLE_CLOUD_PROJECT"] = profile_model.project_id
        wanna_dict.update({"gcp_profile": profile_model})

        wanna_project = WannaProjectModel.model_validate(wanna_dict["wanna_project"])
        wanna_dict.update({"wanna_project": wanna_project})

        # Remove gcp_profiles from the dictionary
        del wanna_dict["gcp_profiles"]

        # Complete validation and metadata enrichment
        wanna_config = WannaConfigModel.model_validate(wanna_dict)

    logger.user_info(f"GCP profile '{profile_model.profile_name}' will be used.")
    logger.user_info(f"Profile details: {profile_model}")

    if gcp_access_allowed:
        # doing this import here speeds up the CLI app considerably
        from google.cloud import aiplatform

        from wanna.core.utils.credentials import get_credentials

        aiplatform.init(
            project=wanna_config.gcp_profile.project_id,
            location=wanna_config.gcp_profile.region,
            credentials=get_credentials(),
        )

    return wanna_config


def load_yaml_maybe_stdin(wanna_config_path: Path) -> dict[str, Any]:
    """
    Loads yaml file from path or stdin if path is '-'

    Args:
        wanna_config_path: path to the yaml file or -

    Returns: loaded yaml

    """
    if str(wanna_config_path) == "-":
        return loaders.load_yaml(
            StringIO(sys.stdin.read()), pathlib.Path(wanna_config_path).parent.resolve()
        )
    else:
        with open(wanna_config_path, encoding="utf-8") as file:
            # Load workflow file
            return loaders.load_yaml(file, pathlib.Path(wanna_config_path).parent.resolve())



import json
from typing import Callable

from wanna.core.deployment.io import IOMixin
from wanna.core.deployment.models import (
    ContainerArtifact,
    JsonArtifact,
    PathArtifact,
    PushResult,
    PushTask,
)
from wanna.core.loggers.wanna_logger import get_logger

logger = get_logger(__name__)


class ArtifactsPushMixin(IOMixin):
    def push_artifacts(
        self, docker_pusher: Callable[[list[str]], None], push_tasks: list[PushTask]
    ) -> PushResult:
        def push_containers(container_artifacts: list[ContainerArtifact]):
            for artifact in container_artifacts:
                with logger.user_spinner(f"Pushing {artifact.name.lower()} to {artifact.tags}"):
                    docker_pusher(artifact.tags)

        def push_manifests(manifest_artifacts: list[PathArtifact]):
            for artifact in manifest_artifacts:
                with logger.user_spinner(
                    f"Pushing {artifact.name.lower()} to {artifact.destination}"
                ):
                    self.upload_file(artifact.source, artifact.destination)

        def push_json(artifacts: list[JsonArtifact]):
            for artifact in artifacts:
                with logger.user_spinner(
                    f"Pushing {artifact.name.lower()} to {artifact.destination}"
                ):
                    self.write(artifact.destination, json.dumps(artifact.json_body))

        results: PushResult = []

        for push_task in push_tasks:
            push_containers(container_artifacts=push_task.container_artifacts)
            push_manifests(manifest_artifacts=push_task.manifest_artifacts)
            push_json(artifacts=push_task.json_artifacts)
            results.append(
                (
                    push_task.container_artifacts,
                    push_task.manifest_artifacts,
                    push_task.json_artifacts,
                )
            )

        return results



import re
from functools import lru_cache
from pathlib import Path
from typing import Any

from gcloud_config_helper import gcloud_config_helper
from google.cloud.storage import Blob

from wanna.core.utils.env import should_validate

if should_validate:
    # since is very slow to import all these, we do it only when validation is required
    from google.cloud.compute_v1 import (
        ImagesClient,
        MachineTypesClient,
        RegionsClient,
        ZonesClient,
    )
    from google.cloud.compute_v1.types import ListImagesRequest

from google.cloud import storage
from google.cloud.resourcemanager_v3.services.projects import ProjectsClient

from wanna.core.utils.credentials import get_credentials

NETWORK_REGEX = (
    "projects/((?:(?:[-a-z0-9]{1,63}\\.)*(?:[a-z](?:[-a-z0-9]{0,61}[a-z0-9])?):)"
    "?(?:[0-9]{1,19}|(?:[a-z0-9](?:[-a-z0-9]{0,61}[a-z0-9])?)))/global/networks/"
    "((?:[a-z](?:[-a-z0-9]*[a-z0-9])?))$"
)


@lru_cache(maxsize=256)
def get_available_compute_machine_types(project_id: str, zone: str) -> list[str]:
    """
    Get available GCP Compute Engine Machine Types based on project and zone.
    Args:
        project_id: GCP project id
        zone: GCP project location (zone)

    Returns:
        list of available machine types
    """
    if should_validate:
        response = MachineTypesClient(credentials=get_credentials()).list(
            project=project_id, zone=zone
        )
        machine_types = [mtype.name for mtype in response.items]
    else:
        machine_types = [
            "c2-standard-4",
            "c2-standard-8",
            "c2-standard-16",
            "c2-standard-30",
            "c2-standard-60",
            "e2-highcpu-2",
            "e2-highcpu-4",
            "e2-highcpu-8",
            "e2-highcpu-16",
            "e2-highcpu-32",
            "e2-highmem-2",
            "e2-highmem-4",
            "e2-highmem-8",
            "e2-highmem-16",
            "e2-medium",
            "e2-micro",
            "e2-small",
            "e2-standard-2",
            "e2-standard-4",
            "e2-standard-8",
            "e2-standard-16",
            "e2-standard-32",
            "f1-micro",
            "g1-small",
            "m1-megamem-96",
            "m1-ultramem-160",
            "m1-ultramem-40",
            "m1-ultramem-80",
            "m2-megamem-416",
            "m2-ultramem-208",
            "m2-ultramem-416",
            "n1-highcpu-2",
            "n1-highcpu-4",
            "n1-highcpu-8",
            "n1-highcpu-16",
            "n1-highcpu-32",
            "n1-highcpu-64",
            "n1-highcpu-96",
            "n1-highmem-16",
            "n1-highmem-2",
            "n1-highmem-4",
            "n1-highmem-8",
            "n1-highmem-32",
            "n1-highmem-64",
            "n1-highmem-96",
            "n1-megamem-96",
            "n1-standard-1",
            "n1-standard-2",
            "n1-standard-4",
            "n1-standard-8",
            "n1-standard-16",
            "n1-standard-32",
            "n1-standard-64",
            "n1-standard-96",
            "n1-ultramem-40",
            "n1-ultramem-80",
            "n1-ultramem-160",
            "n2-highcpu-2",
            "n2-highcpu-4",
            "n2-highcpu-8",
            "n2-highcpu-16",
            "n2-highcpu-32",
            "n2-highcpu-48",
            "n2-highcpu-64",
            "n2-highcpu-80",
            "n2-highcpu-96",
            "n2-highmem-2",
            "n2-highmem-4",
            "n2-highmem-8",
            "n2-highmem-16",
            "n2-highmem-32",
            "n2-highmem-48",
            "n2-highmem-64",
            "n2-highmem-80",
            "n2-highmem-96",
            "n2-highmem-128",
            "n2-standard-2",
            "n2-standard-4",
            "n2-standard-8",
            "n2-standard-16",
            "n2-standard-32",
            "n2-standard-48",
            "n2-standard-64",
            "n2-standard-80",
            "n2-standard-96",
            "n2-standard-128",
            "n2d-highcpu-128",
            "n2d-highcpu-2",
            "n2d-highcpu-4",
            "n2d-highcpu-8",
            "n2d-highcpu-16",
            "n2d-highcpu-32",
            "n2d-highcpu-48",
            "n2d-highcpu-64",
            "n2d-highcpu-80",
            "n2d-highcpu-96",
            "n2d-highcpu-224",
            "n2d-highmem-2",
            "n2d-highmem-4",
            "n2d-highmem-8",
            "n2d-highmem-16",
            "n2d-highmem-32",
            "n2d-highmem-48",
            "n2d-highmem-64",
            "n2d-highmem-80",
            "n2d-highmem-96",
            "n2d-standard-2",
            "n2d-standard-4",
            "n2d-standard-8",
            "n2d-standard-16",
            "n2d-standard-32",
            "n2d-standard-48",
            "n2d-standard-64",
            "n2d-standard-80",
            "n2d-standard-96",
            "n2d-standard-128",
            "n2d-standard-224",
            "t2d-standard-1",
            "t2d-standard-2",
            "t2d-standard-4",
            "t2d-standard-8",
            "t2d-standard-16",
            "t2d-standard-32",
            "t2d-standard-48",
            "t2d-standard-60",
            "a2-highgpu-1g",
            "a2-highgpu-2g",
            "a2-highgpu-4g",
            "a2-highgpu-8g",
            "a2-megagpu-16g",
        ]

    return machine_types


@lru_cache(maxsize=32)
def get_available_zones(project_id: str) -> list[str]:
    """
    Get available GCP zones based on project.
    Args:
        project_id: GCP project id

    Returns:
        list of available zones
    """

    if should_validate:
        response = ZonesClient(credentials=get_credentials()).list(project=project_id)
        return [zone.name for zone in response.items]
    else:
        return [
            "us-central1-a",
            "us-central1-b",
            "us-central1-c",
            "us-central1-f",
            "us-east1-b",
            "us-east1-c",
            "us-east1-d",
            "us-west1-a",
            "us-west1-b",
            "us-west1-c",
            "europe-west1-b",
            "europe-west1-c",
            "europe-west1-d",
            "europe-west3-a",
            "europe-west3-b",
            "europe-west3-c",
            "europe-west4-a",
            "europe-west4-b",
            "europe-west4-c",
        ]


@lru_cache(maxsize=32)
def get_available_regions(project_id: str) -> list[str]:
    """
    Get available GCP regions based on project.
    Args:
        project_id: GCP project id

    Returns:
        list of available regions
    """
    if should_validate:
        response = RegionsClient(credentials=get_credentials()).list(project=project_id)
        return [region.name for region in response.items]
    else:
        return [
            "europe-west1",
            "europe-west3",
            "europe-west4",
            "us-east1",
            "us-west1",
            "us-central1",
        ]


def get_region_from_zone(zone: str) -> str:
    """
    Get available GCP region from zone.
    Args:
        zone: GCP zone

    Returns:
        region: GCP region
    """
    return zone.rpartition("-")[0]


@lru_cache(maxsize=32)
def convert_project_id_to_project_number(project_id: str) -> str:
    """
    Convert GCP project_id (eg. 'my-gcp-project') to project_number (eg. '966193337054')

    Args:
        project_id: GCP project id

    Returns:
        project_number: GCP project number
    """
    project_name = (
        ProjectsClient(credentials=get_credentials())
        .get_project(name=f"projects/{project_id}")
        .name
    )
    project_number = re.sub("projects/", "", project_name)
    return project_number


def parse_image_name_family(name: str) -> dict[str, Any]:
    """
    Based on GCP Compute Engine VM Image name family (eg. tf2-2-7-cu113-notebooks-debian-10)
    return framework (eg. tf2), version (eg. 2-7-cu113), os (debian-10) information.

    Args:
        name: VM Image family name

    Returns:
        Dictionary with framework, version and os

    """
    framework = name.partition("-")[0]
    match_version = re.search("(?<=-)(.*?)(?=-notebooks)", name)
    assert match_version is not None  # TODO: improve result assert with message
    version = match_version.group()
    os = None if name.endswith("notebooks") else name.split("-notebooks-")[-1]
    return {"framework": framework, "version": version, "os": os}


@lru_cache(maxsize=128)
def get_available_compute_image_families(
    project: str,
    image_filter: str | None = None,
    family_must_contain: str | None = None,
) -> list[dict[str, str]]:
    """
    List available Compute Engine VM image families.

    Args:
        project: VM Image project ID
        image_filter: filter for the images https://googleapis.dev/python/compute/latest/compute_v1/types.html#google.cloud.compute_v1.types.ListImagesRequest.filter
        family_must_contain: additional string that must be a part of the image family name for easier filtering
                                (eg. notebook to filter only the Vertex AI Workbench notebook-ready images)

    Returns:
        List of dicts from parse_image_name_family

    """  # noqa: E501
    list_images_request = ListImagesRequest(project=project, filter=image_filter)
    all_images = ImagesClient(credentials=get_credentials()).list(list_images_request)
    if family_must_contain:
        return [
            parse_image_name_family(image.family)
            for image in all_images
            if family_must_contain in image.family
        ]
    return [parse_image_name_family(image.family) for image in all_images]


def upload_file_to_gcs(filename: Path, bucket_name: str, blob_name: str) -> storage.blob.Blob:
    """
    Upload file to GCS bucket

    Args:
        filename: local file
        bucket_name:
        blob_name:

    Returns:
        storage.blob.Blob
    """
    bucket = storage_client().get_bucket(bucket_name)
    blob = bucket.blob(blob_name)
    blob.upload_from_filename(filename)
    return blob


def upload_string_to_gcs(data: str, bucket_name: str, blob_name: str) -> storage.blob.Blob:
    """
    Upload a string to GCS bucket without saving it locally as a file.
    Args:
        data: string that will form a file on GCS
        bucket_name:
        blob_name:

    Returns:
        storage.blob.Blob
    """
    bucket = storage_client().get_bucket(bucket_name)
    blob = bucket.blob(blob_name)
    blob.upload_from_string(data)
    return blob


@lru_cache(maxsize=1)
def storage_client() -> storage.Client:
    return storage.Client(credentials=get_credentials())


def download_script_from_gcs(gcs_path: str) -> str:
    """
    Download a script from GCS bucket and return it as a string

    Args:
        gcs_path: GCS path to the script

    Returns:
        str
    """
    blob = Blob.from_string(gcs_path, storage_client())
    return blob.download_as_string().decode("utf-8")


def is_gcs_path(path: str):
    """
    Simply checks if path str represents a GCS path
    Args:
        path: path string that will be checked

    Returns:
        bool
    """
    return path.startswith("gs://")


def get_network_info(network: str | None) -> tuple[str, str] | None:
    """
    gets information about a network if set in long format
    Args:
        network: string to extract network info from
        bucket_name:
        blob_name:

    Returns:
        tuple[str, str]
    """
    if network:
        result = re.search(NETWORK_REGEX, network)
        if result:
            return result.group(1), result.group(2)

    return None


def verify_gcloud_presence():
    if not gcloud_config_helper.on_path():
        # gcloud is needed in the wanna.core.utils.config_enricher_generate_default_labels
        raise OSError("gcloud is not on the path. Wanna-ml does not work properly without it.")



from pathlib import Path
from typing import Any, TextIO

import yaml
from yamlinclude import YamlIncludeConstructor


def replace_environment_variables() -> None:
    """Enable yaml loader to process the environment variables in the yaml file."""
    import os
    import re

    # eg. ${USER_NAME}, ${PASSWORD}
    env_var_pattern = re.compile(r"^(.*)\$\{(.*)\}(.*)$")
    yaml.add_implicit_resolver("!env_var", env_var_pattern)

    def env_var_constructor(loader: Any, node: Any) -> Any:
        """Process environment variables found in the YAML."""
        value = loader.construct_scalar(node)
        return os.path.expandvars(value)

    yaml.add_constructor("!env_var", env_var_constructor)


def load_yaml(stream: TextIO, context_dir: Path, **extras: Any) -> dict[Any, Any]:
    """
    Convert a YAML stream into a class via the OrderedLoader class.
    """
    YamlIncludeConstructor.add_to_loader_class(loader_class=yaml.FullLoader, base_dir=context_dir)
    replace_environment_variables()
    yaml_dict = yaml.load(stream, Loader=yaml.FullLoader) or {}
    yaml_dict.update(extras)
    return yaml_dict


def load_yaml_path(path: Path, context_dir: Path, **extras: Any) -> dict[Any, Any]:
    """
    Convert a Path into a yaml dict
    """
    with open(path, "r", encoding="utf-8") as f:
        return load_yaml(f, context_dir, **extras)



import os
import tarfile
from pathlib import Path

import igittigitt


def tar_docker_context(source_dir: Path, target_tar_file: Path, ignore_patterns: list[str] = []):
    """
    Tars a directory recursively while optionally skipping files based on ignore patterns.

    :param source_dir: Path to the directory to be tarred.
    :param target_tar_file: Path to the output TAR file.
    :param ignore_patterns: List of file patterns to skip (e.g., ['*.pyc', '*.log']).
    """

    parser = igittigitt.IgnoreParser()
    for pattern in ignore_patterns:
        parser.add_rule(pattern, source_dir)

    os.makedirs(target_tar_file.parent.absolute(), exist_ok=True)
    with tarfile.open(target_tar_file, "w:gz") as the_tar_file:
        for root, _, files in os.walk(source_dir):
            for file in files:
                file_path = os.path.join(root, file)
                if parser.match(file_path):
                    continue
                the_tar_file.add(file_path, arcname=os.path.relpath(file_path, source_dir))



import os

from wanna.core.loggers.wanna_logger import get_logger

logger = get_logger(__name__)


def get_env_bool(value: str | None, fallback: bool) -> bool:
    """
    Checks if a str value can be considered True or False,
        mostly used to extract bool values from env vars

    Returns:
        bool from extracted value
    """
    if value in ["False", "false", "0"]:
        return False
    elif value in ["True", "true", "1"]:
        return True
    else:
        return fallback


def _gcp_access_allowed(env_var="WANNA_GCP_ACCESS_ALLOWED"):
    """
    Based on WANNA_GCP_ACCESS_ALLOWED env var checks if wanna can access GCP
    required for on-prem build environments without access to GCP

    Returns:
        bool if wanna can access GCP apis
    """
    allowed = get_env_bool(os.environ.get(env_var), True)
    logger.user_info(f"GCP access {'NOT ' if not allowed else ''}allowed")
    return allowed


gcp_access_allowed = _gcp_access_allowed()


def _should_validate(env_var="WANNA_GCP_ENABLE_REMOTE_VALIDATION"):
    """
    Based on WANNA_GCP_ENABLE_REMOTE_VALIDATION env var checks if wanna should
    run remote validations that access GCP APIs.

    This is mostly to improve development experience, as it can be slow to allways
    run remote validations

    Returns:
        bool if wanna can access GCP apis
    """
    validate = get_env_bool(os.environ.get(env_var), True)
    if validate:
        logger.user_info("Remote validation is enabled")
    else:
        logger.user_info("Remote validation is disabled")
    return validate


def _cloud_build_access_allowed(env_var="WANNA_GCP_CLOUD_BUILD_ACCESS_ALLOWED") -> bool:
    """
    Based on WANNA_GCP_CLOUD_BUILD_ACCESS_ALLOWED env var checks if wanna should
    overwrite cloud build config in wanna.yaml

    This is required for when CI/CD pipeline infra does not allow to build containers
    on GCP

    Returns:
        bool if wanna builds can access GCP Cloud Build service
    """
    allowed = get_env_bool(os.environ.get(env_var), True)
    if allowed:
        logger.user_info("Cloud build access is enabled")
    else:
        logger.user_info("Cloud build access is disabled")
    return allowed


cloud_build_access_allowed = _cloud_build_access_allowed()
should_validate = gcp_access_allowed and _should_validate()


def reload_setup() -> None:
    """
    Reloads the setup of the env vars
    """
    global gcp_access_allowed, should_validate, cloud_build_access_allowed
    gcp_access_allowed = _gcp_access_allowed()
    should_validate = gcp_access_allowed and _should_validate()
    cloud_build_access_allowed = _cloud_build_access_allowed()



import atexit
import json
import os
import zipfile
from pathlib import Path

from google.api_core.exceptions import AlreadyExists
from google.cloud import logging
from google.cloud.aiplatform import PipelineJob
from google.cloud.aiplatform.compat.types import (
    pipeline_state_v1 as gca_pipeline_state_v1,
)
from google.cloud.functions_v1 import CloudFunctionsServiceClient

from wanna.core.deployment.artifacts_push import ArtifactsPushMixin
from wanna.core.deployment.models import (
    AlertPolicyResource,
    CloudFunctionResource,
    CloudSchedulerResource,
    LogMetricResource,
    NotificationChannelResource,
    PipelineResource,
)
from wanna.core.deployment.vertex_scheduling import VertexSchedulingMixIn
from wanna.core.loggers.wanna_logger import get_logger
from wanna.core.models.notification_channel import (
    EmailNotificationChannel,
    PubSubNotificationChannel,
)
from wanna.core.services.path_utils import PipelinePaths
from wanna.core.utils import templates
from wanna.core.utils.gcp import is_gcs_path
from wanna.core.utils.loaders import load_yaml_path
from wanna.core.utils.time import get_timestamp, update_time_template

logger = get_logger(__name__)


class VertexPipelinesMixInVertex(VertexSchedulingMixIn, ArtifactsPushMixin):
    @staticmethod
    def _at_pipeline_exit(pipeline_name: str, pipeline_job: PipelineJob, sync: bool) -> None:
        @atexit.register
        def stop_pipeline_job():
            if sync and pipeline_job and getattr(pipeline_job._gca_resource, "name", None):
                if (
                    pipeline_job.state
                    != gca_pipeline_state_v1.PipelineState.PIPELINE_STATE_SUCCEEDED
                ):
                    logger.user_error(
                        "detected exit signal, "
                        f"shutting down running pipeline {pipeline_name} "
                        f"at {pipeline_job._dashboard_uri()}."
                    )
                    pipeline_job.wait()
                    pipeline_job.cancel()

    def run_pipeline(
        self,
        resource: PipelineResource,
        extra_params: Path | None,
        sync: bool = True,
    ) -> None:
        mode = "sync mode" if sync else "fire-forget mode"

        logger.user_info(f"Running pipeline {resource.pipeline_name} in {mode}")

        # fetch compiled params
        pipeline_job_id = f"pipeline-{resource.pipeline_name}-{get_timestamp()}"

        # Apply override with cli provided params file
        override_params = load_yaml_path(extra_params, Path(".")) if extra_params else {}
        pipeline_params = {**resource.parameter_values, **override_params}
        pipeline_params = update_time_template(pipeline_params)

        # Define Vertex AI Pipeline job
        pipeline_job = PipelineJob(
            display_name=resource.pipeline_name,
            job_id=pipeline_job_id,
            template_path=str(resource.json_spec_path),
            pipeline_root=resource.pipeline_root,
            parameter_values=pipeline_params,
            enable_caching=resource.enable_caching,
            labels=resource.labels,
            project=resource.project,
            location=resource.location,
            encryption_spec_key_name=resource.encryption_spec_key_name,
        )

        VertexPipelinesMixInVertex._at_pipeline_exit(resource.pipeline_name, pipeline_job, sync)

        # submit pipeline job for execution
        experiment = (
            resource.experiment if resource.experiment else f"{resource.pipeline_name}-experiment"
        )

        pipeline_job.submit(
            service_account=resource.service_account,
            network=resource.network,
            experiment=experiment,
        )

        if sync:
            logger.user_info(f"Pipeline dashboard at {pipeline_job._dashboard_uri()}.")
            pipeline_job.wait()

    def deploy_pipeline(
        self,
        resource: PipelineResource,
        pipeline_paths: PipelinePaths,
        version: str,
        env: str,
    ) -> None:
        schedule = (
            next(
                iter([s for s in resource.schedule if s.environment == env]),
                None,
            )
            if isinstance(resource.schedule, list)
            else resource.schedule
        )

        pipeline_service_account = (
            schedule.service_account
            if schedule and schedule.service_account
            else resource.service_account
        )

        base_resource = {
            "project": resource.project,
            "location": resource.location,
            "service_account": str(pipeline_service_account),
        }

        # Create notification channels
        channels = []
        for config in resource.notification_channels:
            if isinstance(config, EmailNotificationChannel):
                for email in config.emails:
                    channel_config = {"email_address": str(email)}
                    name = email.split("@")[0].replace(".", "-")
                    channel = self.upsert_notification_channel(
                        resource=NotificationChannelResource(
                            type_=config.type,
                            description=config.description,
                            name=f"{name}-wanna-email-channel",
                            config=channel_config,
                            labels=resource.labels,
                            **base_resource,
                        )
                    )
                    channels.append(channel.name)
            elif isinstance(config, PubSubNotificationChannel):
                for topic in config.topics:
                    project_id = resource.compile_env_params.get("project_id")
                    channel_config = {"topic": f"projects/{project_id}/topics/{topic}"}
                    channel = self.upsert_notification_channel(
                        resource=NotificationChannelResource(
                            type_=config.type,
                            description=config.description,
                            name=f"{topic}-wanna-alert-topic-channel",
                            config=channel_config,
                            labels=resource.labels,
                            **base_resource,
                        )
                    )
                    channels.append(channel.name)
            else:
                raise ValueError(
                    f"Validation error notification config {config} can't be handled by wanna-ml"
                )

        function = self.upsert_cloud_function(
            resource=CloudFunctionResource(
                name=resource.pipeline_name,
                build_dir=pipeline_paths.get_local_pipeline_deployment_path(version),
                resource_root=pipeline_paths.get_gcs_pipeline_deployment_path(version),
                resource_function_template="scheduler_cloud_function.py",
                resource_requirements_template="scheduler_cloud_function_requirements.txt",
                template_vars=resource.dict(),
                env_params=resource.compile_env_params,
                labels=resource.labels,
                network=resource.network,
                notification_channels=channels,
                **base_resource,
            ),
            env=env,
            version=version,
        )

        if schedule:
            pipeline_spec_path = pipeline_paths.get_gcs_pipeline_json_spec_path(version)
            body = {
                "pipeline_spec_uri": pipeline_spec_path,
                "parameter_values": resource.parameter_values,
                "enable_caching": resource.enable_caching,
            }  # TODO extend with execution_date(now) ?

            self.upsert_cloud_scheduler(
                function=function,
                resource=CloudSchedulerResource(
                    name=resource.pipeline_name,
                    body=body,
                    cloud_scheduler=schedule,
                    labels=resource.labels,
                    notification_channels=channels,
                    **base_resource,
                ),
                env=env,
                version=version,
            )

        else:
            logger.user_info(
                "Deployment Manifest does not have a schedule set. Skipping Cloud Scheduler sync"
            )

        logging_metric_ref = f"{resource.pipeline_name}-ml-pipeline-error"
        gcp_resource_type = "aiplatform.googleapis.com/PipelineJob"
        self.upsert_log_metric(
            LogMetricResource(
                project=resource.project,
                name=logging_metric_ref,
                location=resource.location,
                filter_=f"""
            resource.type="{gcp_resource_type}"
            AND severity >= WARNING
            AND resource.labels.pipeline_job_id:"{resource.pipeline_name}"
            """,
                description=f"Log metric for {resource.pipeline_name} vertex ai pipeline",
            )
        )
        logging_policy_name = f"{resource.pipeline_name}-{env}-ml-pipeline-alert-policy"
        self.upsert_alert_policy(
            AlertPolicyResource(
                name=logging_policy_name,
                project=resource.project,
                location=resource.location,
                logging_metric_type=logging_metric_ref,
                resource_type=gcp_resource_type,
                display_name=logging_policy_name,
                labels=resource.labels,
                notification_channels=channels,
            )
        )

        if "wanna_sla_hours" in resource.labels:
            self.upsert_sink(resource)
            self.upsert_sla_function(resource, version, env)

    def upsert_sink(self, resource: PipelineResource):
        """Creates a sink to export logs to the given Cloud Storage bucket.
        The filter determines which logs this sink matches and will be exported
        to the destination.
        """
        logging_client = logging.Client()

        destination = "storage.googleapis.com/" + f"{resource.pipeline_bucket}"[5:]
        filter_ = f"""resource.type="aiplatform.googleapis.com/PipelineJob"
        AND jsonPayload.pipelineName="{resource.pipeline_name}"
        AND jsonPayload.state="PIPELINE_STATE_RUNNING"
        """
        sink_name = f"{resource.pipeline_name}-sink-{resource.pipeline_version}"

        sink = logging_client.sink(sink_name, filter_=filter_, destination=destination)

        if sink.exists():
            logger.user_error(f"Sink {sink.name} already exists.")
            return

        sink.create()
        logger.user_info(f"Created sink {sink.name}")

    def upsert_sla_function(self, resource: PipelineResource, version: str, env: str) -> None:
        logger.user_info(
            f"Deploying {resource.pipeline_name} SLA monitoring function with version {version} to env {env}"
        )
        parent = f"projects/{resource.project}/locations/{resource.location}"
        local_functions_package = "sla.zip"
        functions_gcs_path_dir = f"{resource.pipeline_bucket}/wanna-pipelines/{resource.pipeline_name}/deployment/{version}/functions"
        functions_gcs_path = f"{functions_gcs_path_dir}/sla.zip"
        function_name = f"{resource.pipeline_name}-{env}-{version}"
        function_path = f"{parent}/functions/{function_name}"

        cloud_function = templates.render_template(
            Path("sla_cloud_function.py"),
            labels=json.dumps(resource.labels, separators=(",", ":")),
        )

        requirements = templates.render_template(Path("sla_cloud_function_requirements.txt"))

        with zipfile.ZipFile(local_functions_package, "w") as z:
            z.writestr("main.py", cloud_function)
            z.writestr("requirements.txt", requirements)

        if not is_gcs_path(functions_gcs_path_dir):
            os.makedirs(functions_gcs_path_dir, exist_ok=True)

        self.upload_file(str(local_functions_package), functions_gcs_path)

        cf = CloudFunctionsServiceClient(credentials=self.credentials)

        function = {
            "name": function_path,
            "description": f"wanna {resource.pipeline_name} function for {env} pipeline",
            "source_archive_url": functions_gcs_path,
            "entry_point": "main",
            "runtime": "python39",
            "event_trigger": {
                "event_type": "google.storage.object.finalize",
                "resource": f"projects/{resource.project}/buckets/"
                + f"{resource.pipeline_bucket}"[5:],
            },
            "service_account_email": resource.service_account,
            "labels": resource.labels,
        }
        try:
            cf.create_function({"location": parent, "function": function}).result()
        except AlreadyExists:
            logger.user_error(f"Function {function_name} already exists, no need to re-deploy")



import itertools
import subprocess
from pathlib import Path

from google.api_core.operation import Operation
from google.cloud import compute_v1
from google.cloud.notebooks_v2 import (
    AcceleratorConfig,
    BootDisk,
    DataDisk,
    GceSetup,
    GPUDriverConfig,
    NetworkInterface,
)
from google.cloud.notebooks_v2.services.notebook_service import NotebookServiceClient
from google.cloud.notebooks_v2.types import (
    ContainerImage,
    CreateInstanceRequest,
    Instance,
    ServiceAccount,
    VmImage,
)

from wanna.core.deployment.models import PushMode
from wanna.core.loggers.wanna_logger import get_logger
from wanna.core.models.wanna_config import WannaConfigModel
from wanna.core.models.workbench import InstanceModel
from wanna.core.services.docker import DockerService
from wanna.core.services.tensorboard import TensorboardService
from wanna.core.services.workbench import BaseWorkbenchService, Instances
from wanna.core.utils import templates
from wanna.core.utils.config_enricher import email_fixer
from wanna.core.utils.gcp import (
    download_script_from_gcs,
    upload_string_to_gcs,
)

logger = get_logger(__name__)


class WorkbenchInstanceService(BaseWorkbenchService[InstanceModel]):
    def __init__(
        self,
        config: WannaConfigModel,
        workdir: Path,
        owner: str | None = None,
        version: str = "dev",
    ):
        super().__init__(
            instance_type="workbench-instance",
        )
        self.version = version
        self.instances = config.notebooks
        self.wanna_project = config.wanna_project
        self.bucket_name = config.gcp_profile.bucket
        self.notebook_client = NotebookServiceClient()
        self.config = config
        self.docker_service = (
            DockerService(
                docker_model=config.docker,
                gcp_profile=config.gcp_profile,
                version=version,
                work_dir=workdir,
                wanna_project_name=self.wanna_project.name,
            )
            if config.docker
            else None
        )

        self.owner = owner
        self.tensorboard_service = TensorboardService(config=config)

    def _delete_instance_client(self, instance: InstanceModel) -> Operation:
        return self.notebook_client.delete_instance(
            name=f"projects/{instance.project_id}/locations/"
            f"{instance.zone}/instances/{instance.name}"
        )

    def _create_instance_client(self, request: CreateInstanceRequest) -> Operation:
        return self.notebook_client.create_instance(request)

    def workbench_location(self, instance: InstanceModel) -> str:
        return instance.zone

    def _list_running_instances(self, project_id: str, location: str) -> list[str]:
        """
        list all notebooks with given project_id and location.

        Args:
            project_id: GCP project ID
            location: GCP location (zone)

        Returns:
            instance_names: list of the full names on notebook instances (this includes project_id, and zone)

        """
        instances = self.notebook_client.list_instances(
            parent=f"projects/{project_id}/locations/{location}"
        )
        instance_names = [i.name for i in instances.instances]
        return instance_names

    def _instance_exists(self, instance: InstanceModel) -> bool:
        full_instance_name = (
            f"projects/{instance.project_id}/locations/{instance.zone}/instances/{instance.name}"
        )
        return full_instance_name in self._list_running_instances(
            instance.project_id, instance.zone
        )

    def _create_instance_request(
        self,
        instance: InstanceModel,
        deploy: bool = True,
        push_mode: PushMode = PushMode.all,
    ) -> CreateInstanceRequest:
        # Network
        full_network_name = self._get_resource_network(
            project_id=self.config.gcp_profile.project_id,
            push_mode=PushMode.all,
            resource_network=instance.network,
            fallback_project_network=self.config.gcp_profile.network,
            use_project_number=True,
        )
        subnet = instance.subnet if instance.subnet else self.config.gcp_profile.subnet
        full_subnet_name = self._get_resource_subnet(
            full_network_name,
            subnet,
            instance.region,
        )
        network_interfaces = (
            [
                NetworkInterface(
                    network=full_network_name,
                    subnet=full_subnet_name,
                )
            ]
            if instance.network and instance.subnet
            else None
        )
        # GPU
        if gpu := instance.gpu:
            accelerator_configs = [
                AcceleratorConfig(
                    core_count=gpu.count,
                    type_=gpu.accelerator_type,
                )
            ]
            gpu_driver_config = GPUDriverConfig(
                enable_gpu_driver=gpu.install_gpu_driver,
                custom_gpu_driver_path=gpu.custom_gpu_driver_path,
            )
        else:
            accelerator_configs = None
            gpu_driver_config = None

        # Environment
        if docker_image_ref := instance.environment.docker_image_ref:
            if self.docker_service:
                image_url = self.docker_service.build_container_and_get_image_url(
                    docker_image_ref, push_mode=push_mode
                )
                repository = image_url.partition(":")[0]
                tag = image_url.partition(":")[-1]
                container_image = ContainerImage(
                    repository=repository,
                    tag=tag,
                )
                vm_image = None
            else:
                raise Exception("Docker params in wanna-ml config not defined")
        elif vm_i := instance.environment.vm_image:
            vm_image = VmImage(
                project="cloud-notebooks-managed",
                family=f"workbench-instances" if vm_i.version is None else None,
                name=f"workbench-instances-{vm_i.version}" if vm_i.version is not None else None,
            )
            container_image = None
        else:
            raise ValueError(
                "No notebook environment was found. This should not be possible."
                " Something went wrong during model validation"
            )

        # Disks
        disk_encryption = "CMEK" if self.config.gcp_profile.kms_key else None
        kms_key = self.config.gcp_profile.kms_key if self.config.gcp_profile.kms_key else None
        boot_disk = (
            BootDisk(
                disk_size_gb=boot_d.size_gb,
                disk_type=boot_d.disk_type,
                disk_encryption=disk_encryption,
                kms_key=kms_key,
            )
            if (boot_d := instance.boot_disk)
            else None
        )
        data_disks = (
            [
                DataDisk(
                    disk_size_gb=data_d.size_gb,
                    disk_type=data_d.disk_type,
                    disk_encryption=disk_encryption,
                    kms_key=kms_key,
                )
            ]
            if (data_d := instance.data_disk)
            else None
        )

        # service account and instance owners
        service_accounts = [ServiceAccount(email=sa)] if (sa := instance.service_account) else None
        instance_owner = self.owner or instance.owner
        instance_owners = [instance_owner] if instance_owner else None

        # labels and tags
        tags = instance.tags
        labels = {
            "wanna_name": instance.name,
            "wanna_resource": self.instance_type,
        }
        if instance_owner:
            labels["wanna_owner"] = email_fixer(instance_owner)
        if instance.labels:
            labels = {**instance.labels, **labels}

        # post startup script
        if deploy and (
            instance.bucket_mounts
            or instance.tensorboard_ref
            or self.config.gcp_profile.env_vars
            or instance.env_vars
        ):
            script = self._prepare_startup_script(instance)

            blob = upload_string_to_gcs(
                script,
                instance.bucket or self.bucket_name,
                f"notebooks/{instance.name}/startup_script.sh",
            )
            post_startup_script = f"gs://{blob.bucket.name}/{blob.name}"
        elif instance.post_startup_script is not None:
            post_startup_script = instance.post_startup_script
        else:
            post_startup_script = None

        metadata = instance.metadata or {}
        if instance.enable_monitoring:
            enable_monitoring_metadata = {
                "enable-guest-attributes": "TRUE",
                "report-system-health": "TRUE",
                "report-notebook-metrics": "TRUE",
                "install-monitoring-agent": "TRUE",
                "enable-extended-ui": "TRUE",
            }
            metadata = {**metadata, **enable_monitoring_metadata}
        if instance.collaborative:
            # todo: test this
            collaborative_metadata = {"use-collaborative": "TRUE"}
            metadata = {**metadata, **collaborative_metadata}
        # https://cloud.google.com/vertex-ai/docs/workbench/instances/create-dataproc-enabled#terraform
        dataproc_metadata = {"disable-mixer": "false" if instance.enable_dataproc else "true"}
        metadata = {**metadata, **dataproc_metadata}
        # https://cloud.google.com/vertex-ai/docs/workbench/instances/idle-shutdown#terraform
        idle_shutdown_metadata = {
            "idle-timeout-seconds": str(idle_shutdown_timeout * 60)  # we set it in minutes
            if (idle_shutdown_timeout := instance.idle_shutdown_timeout)
            else ""
        }
        metadata = {**metadata, **idle_shutdown_metadata}
        if post_startup_script:
            # https://cloud.google.com/vertex-ai/docs/workbench/instances/manage-metadata
            # https://cloud.google.com/vertex-ai/docs/workbench/instances/create#gcloud
            post_startup_script_metadata = {
                "post-startup-script": post_startup_script,
                "post-startup-script-behavior": instance.post_startup_script_behavior,
            }
            metadata = {**metadata, **post_startup_script_metadata}
        if instance.bucket_mounts and container_image is not None:
            gcsfuse_metadata = {"container-allow-fuse": "true"}
            metadata = {**metadata, **gcsfuse_metadata}
        if instance.environment_auto_upgrade:
            auto_upgrade_metadata = {
                "notebook-upgrade-schedule": instance.environment_auto_upgrade
            }
            metadata = {**metadata, **auto_upgrade_metadata}
        if instance.delete_to_trash:
            delete_to_trash_metadata = {"notebook-enable-delete-to-trash": "true"}
            metadata = {**metadata, **delete_to_trash_metadata}
        report_health_metadata = {
            "report-event-health": "true" if instance.report_health else "false"
        }
        metadata = {**metadata, **report_health_metadata}

        gce_setup = GceSetup(
            machine_type=instance.machine_type,
            accelerator_configs=accelerator_configs,
            service_accounts=service_accounts,
            vm_image=vm_image,
            container_image=container_image,
            boot_disk=boot_disk,
            data_disks=data_disks,
            network_interfaces=network_interfaces,
            disable_public_ip=instance.no_public_ip,
            tags=tags,
            metadata=metadata,
            enable_ip_forwarding=instance.enable_ip_forwarding,
            gpu_driver_config=gpu_driver_config,
        )
        instance_proto = Instance(
            gce_setup=gce_setup,
            instance_owners=instance_owners,
            disable_proxy_access=instance.no_proxy_access,
            labels=labels,
        )

        return CreateInstanceRequest(
            parent=f"projects/{instance.project_id}/locations/{instance.zone}",
            instance_id=instance.name,
            instance=instance_proto,
        )

    def _prepare_startup_script(self, nb_instance: InstanceModel) -> str:
        """
        Prepare the notebook startup script based on the information from notebook.
        This script run at the Compute Engine Instance creation time with a root user.

        Args:
            nb_instance

        Returns:
            startup_script
        """
        env_vars = self.config.gcp_profile.env_vars if self.config.gcp_profile.env_vars else dict()
        if nb_instance.env_vars:
            env_vars = {**env_vars, **nb_instance.env_vars}

        if nb_instance.tensorboard_ref:
            tensorboard_resource_name = (
                self.tensorboard_service.get_or_create_tensorboard_instance_by_name(
                    nb_instance.tensorboard_ref
                )
            )
        else:
            tensorboard_resource_name = None

        if nb_instance.post_startup_script is not None:
            # download the script from the bucket
            user_script = download_script_from_gcs(nb_instance.post_startup_script)
            # removing the shebang
            lines = user_script.split("\n")
            if lines and lines[0].startswith("#!"):
                lines = lines[1:]
            user_script = "\n".join(lines)
        else:
            user_script = None

        startup_script = templates.render_template(
            Path("notebook_startup_script.sh.j2"),
            bucket_mounts=nb_instance.bucket_mounts,
            tensorboard_resource_name=tensorboard_resource_name,
            env_vars=env_vars,
            user_script=user_script,
        )
        return startup_script

    def _client_get_instance(self, instance_id: str) -> Instances:
        return self.notebook_client.get_instance(name=instance_id)

    def _get_jupyterlab_link(self, instance_id: str) -> str:
        instance_info = self.notebook_client.get_instance({"name": instance_id})
        return f"https://{instance_info.proxy_uri}"

    def _ssh(
        self, workbench_instance: InstanceModel, run_in_background: bool, local_port: int
    ) -> None:
        """
        SSH connect to the notebook instance if the instance is already started.

        Args:
            workbench_instance: notebook model representing the instance you want to connect to
            run_in_background: whether to run in the background or in interactive mode
            local_port: jupyter lab will be exposed at this port at localhost

        Returns:

        """
        exists = self._instance_exists(workbench_instance)
        if not exists:
            logger.user_info(
                f"Notebook {workbench_instance.name} is not running, create it first and then ssh connect to it."
            )
            return

        bash_command = f"gcloud compute ssh \
             --project {workbench_instance.project_id} \
             --zone {workbench_instance.zone} \
             --tunnel-through-iap \
              {workbench_instance.name} \
             -- -L 8080:localhost:{local_port}"
        if run_in_background:
            bash_command += " -N -f"

        process = subprocess.Popen(bash_command.split())
        process.communicate()

    def ssh(self, instance_name: str, run_in_background: bool, local_port: int) -> None:
        """
        A wrapper around _ssh method, but this function also verifies if the notebook is
        defined in YAML config and already started.

        Args:
            instance_name: name of the notebook to connect to
            run_in_background: whether to run in the background or in interactive mode
            local_port: jupyter lab will be exposed at this port at localhost

        Returns:

        """
        if instance_name == "all":
            if len(self.config.notebooks) == 1:
                self._ssh(self.config.notebooks[0], run_in_background, local_port)
            elif len(self.config.notebooks) > 1:
                raise ValueError("You can connect to only one notebook at a time.")
            else:
                logger.user_error("No notebook definition found in your YAML config.")
        else:
            if instance_name in [notebook.name for notebook in self.config.notebooks]:
                self._ssh(
                    [
                        notebook
                        for notebook in self.config.notebooks
                        if notebook.name == instance_name
                    ][0],
                    run_in_background,
                    local_port,
                )
            else:
                logger.user_error(f"No notebook {instance_name} found")
                raise ValueError(f"Notebook {instance_name} does not exists in configuration")

    def build(self) -> int:
        for instance in self.instances:
            self._create_instance_request(
                instance=instance, deploy=False, push_mode=PushMode.manifests
            )
        logger.user_success("Notebooks validation OK!")
        return 0

    def push(self, instance_name: str):
        instances = self._filter_instances_by_name(instance_name)

        docker_image_refs = {
            instance.environment.docker_image_ref
            for instance in instances
            if instance.environment.docker_image_ref
        }
        if docker_image_refs:
            if self.docker_service:
                for docker_image_ref in docker_image_refs:
                    image_tag = self.docker_service.get_image(docker_image_ref=docker_image_ref)
                    if image_tag[1]:
                        self.docker_service.push_image(image_tag[1])
            else:
                raise Exception("Docker params in wanna-ml config not defined")

    def _return_diff(self) -> tuple[list[InstanceModel], list[InstanceModel]]:
        """
        Figuring out the diff between GCP and wanna.yaml. lists user-managed notebooks to be deleted and created.
        """
        # We list compute instances and not notebooks, because with notebooks you cannot list instances in all zones.
        # So instead we list all Compute Engine instances with notebook labels
        instance_client = compute_v1.InstancesClient()
        request = compute_v1.AggregatedListInstancesRequest(
            filter=f"(labels.wanna_resource:notebook) (labels.wanna_project:{self.wanna_project.name})"
        )
        request.project = self.config.gcp_profile.project_id
        agg_list = instance_client.aggregated_list(request=request)
        gcp_instances = itertools.chain(
            *[resp.instances for zone, resp in agg_list if resp.instances]
        )

        active_notebooks = [
            InstanceModel.model_validate(
                {
                    "name": i.name,
                    "zone": i.zone.split("/")[-1],
                    "project_id": self.config.gcp_profile.project_id,
                }
            )
            for i in gcp_instances
        ]
        active_notebook_names = [n.name for n in active_notebooks]
        wanna_notebook_names = [n.name for n in self.instances]

        to_be_deleted = []
        to_be_created = []
        """
        Notebooks to be deleted
        """
        for active_notebook in active_notebooks:
            if active_notebook.name not in wanna_notebook_names:
                to_be_deleted.append(active_notebook)
        """
        Notebooks to be created
        """
        for wanna_notebook in self.instances:
            if wanna_notebook.name not in active_notebook_names:
                to_be_created.append(wanna_notebook)
        return to_be_deleted, to_be_created



from typing import Annotated, Any, Literal, Union

from pydantic import BaseModel, ConfigDict, Field, field_validator, model_validator

from wanna.core.models.base_instance import BaseInstanceModel
from wanna.core.models.gcp_components import GPU, Disk


class PythonPackageModel(BaseModel):
    docker_image_ref: str
    package_gcs_uri: str
    module_name: str

    model_config = ConfigDict(extra="forbid")


class ContainerModel(BaseModel):
    docker_image_ref: str
    command: list[str] | None = None

    model_config = ConfigDict(extra="forbid")


class WorkerPoolModel(BaseModel):
    python_package: PythonPackageModel | None = None
    container: ContainerModel | None = None
    args: list[str | float | int] | None = None
    env: dict[str, str] | None = None
    machine_type: str = "n1-standard-4"
    gpu: GPU | None = None
    boot_disk: Disk | None = None
    replica_count: int = 1

    model_config = ConfigDict(extra="forbid")

    @model_validator(mode="before")
    def one_from_python_or_container_spec_must_be_set(cls, values):  # pylint: disable=no-self-argument,no-self-use
        if values.get("python_package") and values.get("container"):
            raise ValueError("Only one of python_package or container can be set")
        if not values.get("python_package") and not values.get("container"):
            raise ValueError("At least one of python_package or container must be set")
        return values


class ReductionServerModel(BaseModel):
    replica_count: int
    machine_type: str
    container_uri: str


class IntegerParameter(BaseModel):
    type: Literal["integer"]
    var_name: str
    min: int
    max: int
    scale: Literal["log", "linear"] = "linear"

    model_config = ConfigDict(extra="forbid")


class DoubleParameter(BaseModel):
    type: Literal["double"]
    var_name: str
    min: float
    max: float
    scale: Literal["log", "linear"] = "linear"

    model_config = ConfigDict(extra="forbid")


class CategoricalParameter(BaseModel):
    type: Literal["categorical"]
    var_name: str
    values: list[str]

    model_config = ConfigDict(extra="forbid")


class DiscreteParameter(BaseModel):
    type: Literal["discrete"]
    var_name: str
    scale: Literal["log", "linear"] = "linear"
    values: list[int]

    model_config = ConfigDict(extra="forbid")


HyperParamater = Annotated[
    Union[IntegerParameter, DoubleParameter, CategoricalParameter, DiscreteParameter],
    Field(discriminator="type"),
]


class HyperparameterTuning(BaseModel):
    """
    - `metrics` - Dictionary of type [str, Literal["minimize", "maximize"]]
    - `parameters` - list[HyperParamater] defined per var_name, type, min, max, scale
    - `max_trial_count` - [int] defaults to 15
    - `parallel_trial_count` - [int] defaults to 3
    - `search_algorithm` - [str] (optional) Can be "grid" or "random"
    - `encryption_spec` - [str] (optional) The Cloud KMS resource identifier. Has the form:
    projects/my-project/locations/my-region/keyRings/my-kr/cryptoKeys/my-key
    The key needs to be in the same region as where the compute resource is created
    """

    metrics: dict[str, Literal["minimize", "maximize"]]
    parameters: list[HyperParamater]
    max_trial_count: int = 15
    parallel_trial_count: int = 3
    search_algorithm: Literal["grid", "random"] | None = None
    encryption_spec: str | None = None


class BaseCustomJobModel(BaseInstanceModel):
    """
    - `name` - [str] Custom name for this instance
    - `project_id' - [str] (optional) Overrides GCP Project ID from the `gcp_profile` segment
    - `zone` - [str] (optional) Overrides zone from the `gcp_profile` segment
    - `region` - [str] (optional) Overrides region from the `gcp_profile` segment
    - `labels`- [dict[str, str]] (optional) Custom labels to apply to this instance
    - `service_account` - [str] (optional) Overrides service account from the `gcp_profile` segment
    - `network` - [str] (optional) Overrides network from the `gcp_profile` segment
    - `tags`- [dict[str, str]] (optional) Tags to apply to this instance
    - `metadata`- [str] (optional) Custom metadata to apply to this instance
    - `enable_web_access` - [bool] Whether you want Vertex AI to enable interactive shell access
    to training containers. Default is False
    - `bucket` - [str] Overrides bucket from the `gcp_profile` segment
    - `base_output_directory` - [str] (optional) Path to where outputs will be saved
    - `tensorboard_ref` - [str] (optional) Name of the Vertex AI Experiment
    - `timeout_seconds` - [int] Job timeout. Defaults to 60 * 60 * 24 s = 24 hours
    - `encryption_spec`- [str] (optional) The Cloud KMS resource identifier. Has the form:
    projects/my-project/locations/my-region/keyRings/my-kr/cryptoKeys/my-key
    The key needs to be in the same region as where the compute resource is created
    - `env_vars` - dict[str, str] (optional) Environment variables to be propagated to the job
    """

    region: str
    enable_web_access: bool = False
    bucket: str
    base_output_directory: str | None = None
    tensorboard_ref: str | None = None
    timeout_seconds: int = 60 * 60 * 24  # 24 hours
    encryption_spec: Any | None = None
    env_vars: dict[str, str] | None = None

    @model_validator(mode="before")
    def _set_base_output_directory_if_not_provided(  # pylint: disable=no-self-argument,no-self-use
        cls, values: dict[str, Any]
    ) -> dict[str, Any]:
        if not values.get("base_output_directory"):
            values["base_output_directory"] = (
                f"gs://{values.get('bucket')}/wanna-jobs/{values.get('name')}/outputs"
            )
        return values

    @model_validator(mode="before")
    def _service_account_must_be_set_when_using_tensorboard(  # pylint: disable=no-self-argument,no-self-use
        cls, values: dict[str, Any]
    ) -> dict[str, Any]:
        if values.get("tensorboard_ref") and not values.get("service_account"):
            raise ValueError("service_account must be set when using tensorboard in jobs")
        return values


# https://cloud.google.com/vertex-ai/docs/training/create-custom-job
class CustomJobModel(BaseCustomJobModel):
    workers: list[WorkerPoolModel]
    hp_tuning: HyperparameterTuning | None = None

    @field_validator("workers", mode="after")
    def _worker_pool_must_have_same_spec(  # pylint: disable=no-self-argument,no-self-use
        cls, workers: list[WorkerPoolModel]
    ) -> list[WorkerPoolModel]:
        if workers:
            python_packages = list(filter(lambda w: w.python_package is not None, workers))
            containers = list(filter(lambda w: w.container is not None, workers))
            if len(python_packages) > 0 and len(containers) > 0:
                raise ValueError(
                    "CustomJobs must be of the same spec. "
                    "Either just based on python_package or container"
                )

        return workers


# https://cloud.google.com/vertex-ai/docs/training/create-training-pipeline
class TrainingCustomJobModel(BaseCustomJobModel):
    worker: WorkerPoolModel
    reduction_server: ReductionServerModel | None = None


JobModelTypeAlias = Union[CustomJobModel, TrainingCustomJobModel]



import importlib.metadata
import json

import requests

from wanna.core.loggers.wanna_logger import get_logger

logger = get_logger(__name__)

PYPI_URL = "https://pypi.org/pypi/wanna-ml/json"
UPDATE_MESSAGE = (
    "If you used `pipx` to install WANNA CLI, use the following command:\n\n"
    "pipx upgrade wanna-ml\n\n"
    "Otherwise, use `pip install --upgrade wanna-ml`"
    "(exact command will depend on your environment).\n\n"
)


def get_latest_version() -> str:
    try:
        resp = requests.get(PYPI_URL)
        resp.raise_for_status()
    except (ConnectionError, requests.exceptions.RequestException):
        logger.exception("Failed to retrieve versions")
        return ""
    data = json.loads(resp.text)
    return data["info"]["version"]


def perform_check() -> None:
    """Perform the version check and instructs the user about the next steps"""

    latest_version = get_latest_version()
    version = importlib.metadata.version("wanna-ml")
    if latest_version and version < latest_version:
        logger.user_error(
            f"Installed version is {version}, the latest version is {latest_version}",
        )
        logger.user_info(
            UPDATE_MESSAGE,
        )
    else:
        logger.user_success(f"Your wanna cli is up to date with {latest_version}")



from typing import Dict, Optional

from kfp.v2 import dsl


@dsl.component(
    base_image="python:3.9",
    packages_to_install=[
        "google-cloud-pipeline-components==2.5.0",
        "google-cloud-aiplatform==1.35.0",
    ],
)
def get_or_create_endpoint(
    project: str,
    location: str,
    display_name: str,
    endpoint: dsl.Output[dsl.Artifact],
    labels: Dict[str, str] = {},
    network: Optional[str] = None,
):
    import logging

    from google.api_core.client_options import ClientOptions
    from google.cloud import aiplatform_v1

    api_endpoint = f"{location}-aiplatform.googleapis.com"
    vertex_uri_prefix = f"https://{api_endpoint}/v1/"

    client = aiplatform_v1.EndpointServiceClient(
        client_options=ClientOptions(
            api_endpoint=api_endpoint,
        )
    )

    list_request = aiplatform_v1.ListEndpointsRequest(
        parent=f"projects/{project}/locations/{location}",
    )
    resp = client.list_endpoints(request=list_request)
    # Match endpoints on display name and also if they are public / private
    endpoints = [
        e for e in resp if e.display_name == display_name and e.network == (network or "")
    ]

    if len(endpoints) > 0:
        logging.info("Already existing endpoints found")
        # If multiple endpoints are matched, take the newest one by update time
        sorted_endpoints = sorted(endpoints, key=lambda ep: ep.update_time, reverse=True)

        endpoint_resource_name = sorted_endpoints[0].name

    else:
        logging.info("No existing endpoint found, we will create one")
        endpoint_model = aiplatform_v1.Endpoint()
        endpoint_model.display_name = display_name
        endpoint_model.labels = labels
        if network:
            endpoint_model.network = network
        request = aiplatform_v1.CreateEndpointRequest(
            parent=f"projects/{project}/locations/{location}",
            endpoint=endpoint_model,
        )
        operation = client.create_endpoint(request=request)
        response = operation.result()

        endpoint_resource_name = response.name

    endpoint.uri = vertex_uri_prefix + endpoint_resource_name
    endpoint.metadata = {"resourceName": endpoint_resource_name}



# mypy: allow-untyped-calls
from importlib import metadata as importlib_metadata

try:
    __version__ = importlib_metadata.version(__name__)
except importlib_metadata.PackageNotFoundError:
    __version__ = "snapshot"



from pathlib import Path
from typing import Annotated, Any

from pydantic import Field, StringConstraints, model_validator

from wanna.core.models.base_instance import BaseInstanceModel
from wanna.core.models.cloud_scheduler import CloudSchedulerModel, EnvCloudSchedulerModel

PipelineName = Annotated[
    str,
    StringConstraints(
        min_length=3, max_length=63, pattern="^[a-z][a-z0-9-]*[a-z0-9]$", to_lower=True
    ),
]


class PipelineModel(BaseInstanceModel):
    """
    - `name`- [str] Custom name for this instance
    - `project_id' - [str] (optional) Overrides GCP Project ID from the `gcp_profile` segment
    - `zone` - [str] (optional) Overrides zone from the `gcp_profile` segment
    - `region` - [str] (optional) Overrides region from the `gcp_profile` segment
    - `labels`- [dict[str, str]] (optional) Custom labels to apply to this instance
    - `service_account` - [str] (optional) Overrides service account from the `gcp_profile` segment
    - `network` - [str] (optional) Overrides network from the `gcp_profile` segment
    - `tags`- [dict[str, str]] (optional) Tags to apply to this instance
    - `metadata`- [str] (optional) Custom metadata to apply to this instance
    - `pipeline_function` - [str] (optional) Path to a cloud function
    - `pipeline_params` - [str] (optional) Path to params.yaml file
    - `docker_image_ref` - [list[str]] - list of names of docker images
    - `schedule` - [str] (optional) - Scheduler using a cron syntax
    - `tensorboard_ref` - [str] (optional) - Name of the Vertex AI Experiment
    - `notification_channels_ref` - [list[str]] (optional) list of names of notificartins channel described
    by a model: (name: str, type: Literal["email"], emails: list[EmailStr])
    - `sla_hours` - [float] (optional) Time after which the running pipeline gets stopped
    - `enable_caching` - [bool] enable KubeFlow pipeline execution caching
    """

    name: PipelineName
    zone: str
    pipeline_function: str
    pipeline_params: Path | dict[str, Any] | None = None
    docker_image_ref: list[str] = Field(default_factory=list)
    schedule: CloudSchedulerModel | list[EnvCloudSchedulerModel] | None = None
    tensorboard_ref: str | None = None
    network: str | None = None
    notification_channels_ref: list[str] = Field(default_factory=list)
    sla_hours: float | None = None
    enable_caching: bool = True
    experiment: str | None = None

    @model_validator(mode="before")
    def set_experiment(cls, values):  # pylint: disable=no-self-argument,no-self-use
        """
        Set default pipeline experiment name based on pipeline name
        """
        experiment = values.get("experiment")
        name = values.get("name")
        if not experiment and name:
            values["experiment"] = f"{name}-experiment"

        return values



from pydantic import BaseModel, ConfigDict, field_validator, model_validator

from wanna.core.utils import validators
from wanna.core.utils.gcp import get_region_from_zone


class GCPProfileModel(BaseModel):
    """
    `wanna_profile` section of the yaml config consists of the following inputs:

    - `profile_name` - [str] name of the WANNA GCP Profile, `default` will be used if not specified otherwise.
      You could use also for example `dev`, `prod`, or any other string.
    - `project_id` - [str] GCP project id.
    - `zone` - [str] (optional) GCP location zone.
    - `region` - [str] (optional) GCP location region. If the zone is set and the region not, we automatically
      parse the region from the zone (e.g., zone `us-east1-c` automatically sets region `us-east1` if the region
      is not supplied by the user).
    - `labels` - [dict[str, str]] (optional) GCP resource labels that will be added to all resources you create
      with this profile. By default, we also add a few labels based on `wanna_project` section.
    - `bucket` - [str] (optional) GCS Bucket that can later be used in uploading manifests, storing logs, etc.
      depending on the resource type.
    - `service_account` - [str] (optional) GCP service account that will be used by the created resources.
    If not specified, usually the default service account for each resource type is used.
    - `network` - [str] Google Cloud VPC network name
    - `subnet` - [str] (optional) Google Cloud VPC subnetwork name
    - `kms_key` - [str] (optional) Customer managed enryption key given in format
      projects/{project_id}/locations/{region}/keyRings/{key_ring_id}/cryptoKeys/{key_id}
      If you get an error, please grant the Service Account with the Cloud KMS CryptoKey Encrypter/Decrypter role
    - `docker_repository` - [str] Wanna Docker Repository
    - `docker_registry` - [str] (optional) Wanna Docker Registry, usually in format {region}-docker.pkg.dev
    - `env_vars` - dict[str, str] (optional) Environment variables to be propagated to all the notebooks and custom jobs
    """

    profile_name: str
    project_id: str
    zone: str | None = None
    region: str
    labels: dict[str, str] | None = None
    bucket: str
    service_account: str | None = None
    network: str | None = None
    subnet: str | None = None
    kms_key: str | None = None
    docker_repository: str = "wanna"
    docker_registry: str | None = None
    env_vars: dict[str, str] | None = None

    model_config = ConfigDict(extra="forbid")

    _project_id = field_validator("project_id")(validators.validate_project_id)
    _zone = field_validator("zone")(validators.validate_zone)
    _labels = field_validator("labels")(validators.validate_labels)
    _region = field_validator("region")(validators.validate_region)

    @model_validator(mode="before")
    def parse_region_from_zone(cls, values):  # pylint: disable=no-self-argument,no-self-use
        """
        In some cases, the zone is defined and region not.
        Region can be easily parsed from zone.
        """
        zone, region = (
            values.get("zone"),
            values.get("region"),
        )
        if (region is None) and (zone is not None):
            values["region"] = get_region_from_zone(zone)
        return values



import logging
from typing import cast

import typer
from google.cloud import aiplatform
from google.cloud.aiplatform.tensorboard.tensorboard_resource import (
    Tensorboard,
    TensorboardExperiment,
)
from treelib import Tree

from wanna.core.loggers.wanna_logger import get_logger
from wanna.core.models.tensorboard import TensorboardModel
from wanna.core.models.wanna_config import WannaConfigModel
from wanna.core.services.base import BaseService

logger_gcs = logging.getLogger("google.cloud")
logger_gcs.setLevel(logging.ERROR)

logger = get_logger(__name__)


class TensorboardService(BaseService[TensorboardModel]):
    def __init__(self, config: WannaConfigModel):
        super().__init__(
            instance_type="tensorboard",
        )
        self.config = config
        self.instances = config.tensorboards

    def _delete_one_instance(self, instance: TensorboardModel) -> None:
        """
        Deletes one running Tensorboard Instance.

        Args:
            instance:
        """
        tensorboard = self._find_existing_tensorboard_by_model(instance)
        if not tensorboard:
            logger.user_info(f"Tensorboard {instance.name} does not exist, nothing to delete.")
        else:
            with logger.user_spinner(f"Deleting Tensorboard {instance.name}"):
                aiplatform.Tensorboard(tensorboard_name=tensorboard.resource_name).delete()

    def _create_one_instance(
        self,
        instance: TensorboardModel,
        **kwargs,  # noqa: ARG002
    ) -> None:
        """
        Creates one Tensorboard instance based on pydantic model.

        Args:
            instance:

        Returns:

        """
        if self._instance_exists(instance):
            existing_instance = self._find_existing_tensorboard_by_model(instance)
            if existing_instance:
                logger.user_info(
                    f"Tensorboard {instance.name} already exists and is running at {existing_instance.resource_name}"
                )
                should_recreate = typer.confirm(
                    "Are you sure you want to delete it and start a new?"
                )
                if should_recreate:
                    self._delete_one_instance(instance)
                else:
                    return
        with logger.user_spinner(f"Creating Tensorboard {instance.name}"):
            aiplatform.Tensorboard.create(
                display_name=instance.name,
                description=instance.description,
                labels=instance.labels,
                project=instance.project_id,
                location=instance.region,
            )

        created = self._find_existing_tensorboard_by_model(instance)
        if created:
            logger.user_info(f"Tensorboard {instance.name} is running at {created.resource_name}")

    def _find_existing_tensorboard_by_model(
        self, instance: TensorboardModel
    ) -> Tensorboard | None:
        """
        Given pydantic tensorboard model, find the actual running tensorboard instance on GCP.

        Args:
            instance:

        Returns:
            Tensorboard or None if not found
        """
        for running_tensorboard in self._list_running_instances(
            instance.project_id, instance.region
        ):
            if running_tensorboard.display_name == instance.name:
                return running_tensorboard
        return None

    def _instance_exists(self, instance: TensorboardModel) -> bool:
        """
        Find if there is any running tensorboard instance on GCP
        that corresponds to pydantic TensorboardModel.
        Args:
            instance:

        Returns:
            True is we find a match, False otherwise
        """
        found = self._find_existing_tensorboard_by_model(instance)
        return found is not None

    @staticmethod
    def _list_running_instances(project_id: str, region: str) -> list["Tensorboard"]:
        """
        list all tensorboards with given project_id and region

        Args:
            project_id: GCP project ID
            region: GCP region

        Returns:
            instances: list of the tensorboard instances

        """
        instances = cast(
            list[Tensorboard],
            aiplatform.Tensorboard.list(project=project_id, location=region),
        )
        return instances

    def _find_tensorboard_model_by_name(self, tb_name: str) -> TensorboardModel:
        """
        Finds tensorboard model when given the name of the model.
        Args:
            tb_name: name of the tensorboard model

        Returns:
            TensorboardModel
        """
        matched_tb_models: list[TensorboardModel] = list(
            filter(lambda i: i.name.strip() == tb_name.strip(), self.instances)
        )

        if len(matched_tb_models) == 0:
            raise ValueError(f"No tensorboard model with name {tb_name} found")
        elif len(matched_tb_models) > 1:
            raise ValueError(
                f"Multiple tensorboard models with name {tb_name} found, please use unique names"
            )
        else:
            return matched_tb_models[0]

    def get_or_create_tensorboard_instance_by_name(self, tensorboard_name: str) -> str:
        """
        Given the name of the tensorboard model, it will either return the full resource name of
        existing tensborbord (eg. projects/966197297054/locations/europe-west4/tensorboards/498421815010394112)
        or create a tensorboard with given name if it doesnt exist yet and then return the full path
        Args:
            tensorboard_name:

        Returns:
            full tensorboard resource name
        """
        tb_model = self._find_tensorboard_model_by_name(tb_name=tensorboard_name)
        tb_existing = self._find_existing_tensorboard_by_model(instance=tb_model)
        if not tb_existing:
            logger.user_info(
                f"Tensorboard with name {tb_model.name} in {tb_model.region} not found, creating it."
            )
            self._create_one_instance(tb_model)
            tb_existing = self._find_existing_tensorboard_by_model(instance=tb_model)
            if not tb_existing:
                raise ValueError("Error when creating Tensorboard instance")
        return tb_existing.resource_name

    @staticmethod
    def construct_tb_experiment_url_link(experiment: TensorboardExperiment) -> str:
        """
        Google API doesnt provide any support for getting a URL link to tensorboard experiment.
        Hence use this helper function to do it.
        Args:
            experiment:

        Returns:
            link to the experiment
        """
        return (
            f"https://{experiment.location}.tensorboard.googleusercontent.com/"
            f'experiment/{experiment.resource_name.replace("/", "+")}'
        )

    def _create_tensorboard_tree(self, region: str, filter_expr: str, show_url: bool) -> Tree:
        """
        Create a tensorboard instance - tensorboard experiment - tensorboard run tree
        Args:
            region: gcp region
            filter_expr: gcp filter expression
            show_url: wheather to show url to experiments

        Returns:
            tree
        """
        tree = Tree()
        project_id = self.config.gcp_profile.project_id
        root_tag = f"{project_id} / {region}"
        tree.create_node(tag=root_tag, identifier=root_tag)

        tensorboards = aiplatform.Tensorboard.list(
            project=project_id, location=region, filter=filter_expr
        )
        for tensorboard in tensorboards:
            tag = f"Tensorboard: {tensorboard.display_name}"
            tree.create_node(
                tag=tag,
                identifier=tensorboard.resource_name,
                parent=root_tag,
                data=tensorboard,
            )
            experiments = aiplatform.TensorboardExperiment.list(tensorboard.resource_name)
            for experiment in experiments:
                tag = f"Experiment: {experiment.display_name or experiment.name}"
                if show_url:
                    tag += " " + self.construct_tb_experiment_url_link(experiment)
                tree.create_node(
                    tag=tag,
                    identifier=experiment.resource_name,
                    parent=tensorboard.resource_name,
                    data=experiment,
                )
                runs = aiplatform.TensorboardRun.list(
                    tensorboard_experiment_name=experiment.resource_name
                )
                for run in runs:
                    tag = f"Run: {run.display_name or run.name}"
                    tree.create_node(
                        tag=tag,
                        identifier=run.resource_name,
                        parent=experiment.resource_name,
                        data=run,
                    )
        return tree

    def list_tensorboards_in_tree(self, region: str, filter_expr: str, show_url: bool) -> None:
        with logger.user_spinner("Creating Tensorboard tree"):
            tree = self._create_tensorboard_tree(
                region=region, filter_expr=filter_expr, show_url=show_url
            )
        tree.show()



from typing import Dict, List, Optional

from google_cloud_pipeline_components.types.artifact_types import VertexModel
from kfp import dsl


@dsl.component(
    base_image="python:3.9",
    packages_to_install=[
        "google-cloud-pipeline-components==2.3.0",
        "google-cloud-aiplatform==1.31.0",
    ],
)
def upload_model_version(
    project: str,
    location: str,
    display_name: str,
    serving_container_image_uri: str,
    artifact_uri: str,
    model: dsl.Output[VertexModel],
    model_output_path: dsl.OutputPath(str),  # type: ignore
    labels: Dict[str, str] = {},
    version_aliases: List[str] = [],
    metadata: List[str] = [],
    model_description: Optional[str] = None,
    version_description: Optional[str] = None,
):
    """
    Creates a new model in Vertex-AI Model Registry. If multiple models with same display_name
    already exist, it will take the one with latest update_time and create only new version.
    """
    import json
    import logging

    from google.api_core.client_options import ClientOptions
    from google.cloud import aiplatform_v1

    api_endpoint = f"{location}-aiplatform.googleapis.com"
    vertex_uri_prefix = f"https://{api_endpoint}/v1/"

    client = aiplatform_v1.ModelServiceClient(
        client_options=ClientOptions(
            api_endpoint=api_endpoint,
        )
    )

    parent = f"projects/{project}/locations/{location}"
    list_request = aiplatform_v1.ListModelsRequest(
        parent=parent, filter=f'displayName="{display_name}"'
    )
    resp = client.list_models(request=list_request)
    models = resp.models

    if len(models) > 0:
        logging.info(
            "Already existing model with same display name found, will create a new version"
        )
        # If multiple models are matched, take the newest one by update time
        sorted_models = sorted(models, key=lambda model: model.update_time, reverse=True)
        parent_model_resource_name = sorted_models[0].name

    else:
        logging.info("No existing model with this display name found")
        parent_model_resource_name = None

    vertex_model = aiplatform_v1.Model()
    vertex_model.display_name = display_name
    vertex_model.container_spec = aiplatform_v1.types.ModelContainerSpec()
    vertex_model.container_spec.image_uri = serving_container_image_uri
    vertex_model.artifact_uri = artifact_uri
    if version_aliases:
        vertex_model.version_aliases = version_aliases
    if model_description:
        vertex_model.description = model_description
    if version_description:
        vertex_model.version_description = version_description
    vertex_model.labels = labels
    vertex_model.metadata = metadata

    request = aiplatform_v1.UploadModelRequest(
        parent_model=parent_model_resource_name,
        parent=parent,
        model=vertex_model,
    )
    operation = client.upload_model(request=request)
    response = operation.result()
    new_model_resource_name = response.model

    model.uri = vertex_uri_prefix + new_model_resource_name
    model.metadata = {"resourceName": new_model_resource_name}

    with open(model_output_path, "w", encoding="utf-8") as output_file:
        output_file.write(json.dumps(model.__dict__))
